# pyOCD debugger
# Copyright (c) 2019-2020 Arm Limited
# Copyright (c) 2021-2023 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
import cmsis_pack_manager
import zipfile
from xml.etree import ElementTree
from pathlib import Path
from unittest.mock import MagicMock

from pyocd.target.pack import (cmsis_pack, flash_algo, pack_target)
from pyocd.target.pack.flm_region_builder import FlmFlashRegionBuilder
from pyocd.target import TARGET
from pyocd.core import memory_map
from pyocd.utility.mask import align_down
from pyocd.coresight.ap import APv1Address

K64F = "MK64FN1M0VDC12"
NRF5340 = "nRF5340_xxAA"
STM32L4R5 = "STM32L4R5AGIx"
LPC55S36 = "LPC55S36JBD100"

TEST_DATA_DIR = Path(__file__).resolve().parents[1] / "data" / "packs"
K64F_PACK_PATH = TEST_DATA_DIR / "NXP.MK64F12_DFP.11.0.0.pack"
K64F_1M0_FLM = "arm/MK_P1M0.FLM"
STM32F4_2M0_FLM = TEST_DATA_DIR / "STM32F4xx_2048.FLM"
NRF5340_APP_FLM = TEST_DATA_DIR / "nrf53xx_application.flm"
NRF_PDSC_PATH = TEST_DATA_DIR / "NordicSemiconductor.nRF_DeviceFamilyPack.8.38.0.pdsc"
STM32L4_PDSC_PATH = TEST_DATA_DIR / "Keil.STM32L4xx_DFP.2.5.0.pdsc"
TEST1_PDSC_PATH = TEST_DATA_DIR / "Test1.pdsc"
TEST2_PDSC_PATH = TEST_DATA_DIR / "Test2_algo_overlaps_alias.pdsc"
LPC55S36_PDSC_PATH = TEST_DATA_DIR / "NXP.LPC55S36_DFP.13.0.0.pdsc"

@pytest.fixture(scope='module')
def pack_ref():
    return cmsis_pack_manager.CmsisPackRef(
                "NXP",
                "MK64F12_DFP",
                "11.0.1",
            )

@pytest.fixture(scope='module')#, autouse=True)
def cache(tmpdir_factory, pack_ref):
    tmp_path = str(tmpdir_factory.mktemp("cpm"))
    c = cmsis_pack_manager.Cache(False, False, json_path=tmp_path, data_path=tmp_path)
    c.download_pack_list([pack_ref])
    return c

@pytest.fixture(scope='module')
def k64dev(cache):
    devs = pack_target.ManagedPacks.get_installed_targets()
    return [d for d in devs if d.part_number == K64F].pop()

@pytest.fixture()#autouse=True)
def fixed_installed_packs(monkeypatch, pack_ref):
    def my_get_installed_packs(cache=None):
        return [pack_ref]
    monkeypatch.setattr(pack_target.ManagedPacks,  'get_installed_packs', my_get_installed_packs)

@pytest.fixture(scope='function')
def k64pack():
    return cmsis_pack.CmsisPack(K64F_PACK_PATH)

@pytest.fixture(scope='function')
def k64f1m0(k64pack):
    return [d for d in k64pack.devices if d.part_number == "MK64FN1M0VLL12"].pop()

@pytest.fixture(scope='function')
def k64algo(k64pack):
    flm = k64pack.get_file(K64F_1M0_FLM)
    return flash_algo.PackFlashAlgo(flm)

@pytest.fixture(scope='function')
def nrf5340appflm():
    return flash_algo.PackFlashAlgo(open(NRF5340_APP_FLM, 'rb'))

@pytest.fixture(scope='function')
def stm32f42mflm():
    return flash_algo.PackFlashAlgo(open(STM32F4_2M0_FLM, 'rb'))

# Replacement for CmsisPackDevice._load_flash_algo() that loads the FLM from the test data dir
# instead of the (unset) CmsisPack object.
def load_test_flm(filename):
    p = TEST_DATA_DIR / Path(filename).name
    return p.open('rb')

@pytest.fixture(scope='function')
def nrfpdsc():
    return cmsis_pack.CmsisPackDescription(None, open(NRF_PDSC_PATH, 'rb')) # type:ignore

@pytest.fixture(scope='function')
def test2pdsc():
    return cmsis_pack.CmsisPackDescription(None, open(TEST2_PDSC_PATH, 'rb')) # type:ignore

@pytest.fixture(scope='function')
def test2dev(test2pdsc):
    dev = test2pdsc.devices[0]
    dev._get_pack_file_cb = load_test_flm
    return dev

@pytest.fixture(scope='function')
def test1pdsc():
    return cmsis_pack.CmsisPackDescription(None, TEST1_PDSC_PATH)

@pytest.fixture(scope='function')
def test1dev(test1pdsc):
    return test1pdsc.devices[0]

# Fixture to provide nRF5340 CmsisPackDevice modified to load FLM from test data dir.
@pytest.fixture(scope='function')
def nrf5340(nrfpdsc):
    dev = [d for d in nrfpdsc.devices if d.part_number == NRF5340].pop()
    dev._get_pack_file_cb = load_test_flm
    return dev

@pytest.fixture(scope='function')
def stm32l4pdsc():
    return cmsis_pack.CmsisPackDescription(None, open(STM32L4_PDSC_PATH, 'rb')) # type:ignore

# Fixture to provide STM32L4R5 CmsisPackDevice modified to load FLM from test data dir.
@pytest.fixture(scope='function')
def stm32l4r5(stm32l4pdsc):
    dev = [d for d in stm32l4pdsc.devices if d.part_number == STM32L4R5].pop()
    dev._get_pack_file_cb = load_test_flm
    return dev

@pytest.fixture(scope='function')
def lpc55s36pdsc():
    return cmsis_pack.CmsisPackDescription(None, open(LPC55S36_PDSC_PATH, 'rb')) # type:ignore

# Fixture to provide STM32L4R5 CmsisPackDevice modified to load FLM from test data dir.
@pytest.fixture(scope='function')
def lpc55s36(lpc55s36pdsc):
    dev = [d for d in lpc55s36pdsc.devices if d.part_number == LPC55S36].pop()
    dev._get_pack_file_cb = load_test_flm
    return dev

# Tests for managed packs. Currently disabled as they fail on most systems.
class Disabled_TestPack:
    def test_get_installed(self, pack_ref):
        p = pack_target.ManagedPacks.get_installed_packs()
        assert p == [pack_ref]

    def test_get_targets(self, k64dev):
        assert k64dev.part_number == K64F

    def test_pop_managed_k64(self):
        pack_target.ManagedPacks.populate_target(K64F)
        assert K64F.lower() in TARGET

    def test_k64_mem_map(self, k64dev):
        map = k64dev.memory_map
        raml = map.get_region_for_address(0x1fff0000)
        ramu = map.get_region_for_address(0x20000000)
        flash = map.get_default_region_of_type(memory_map.MemoryType.FLASH)
        assert raml.start == 0x1fff0000 and raml.length == 0x10000
        assert ramu.start == 0x20000000 and ramu.length == 0x30000
        assert flash.start == 0 and flash.length == 0x100000
        assert flash.sector_size == 0x1000

class TestPack:
    def test_devices(self, k64pack):
        devs = k64pack.devices
        pns = [x.part_number for x in devs]
        assert "MK64FN1M0xxx12" in pns
        assert "MK64FX512xxx12" in pns

    # Make sure CmsisPack can open a zip file too.
    def test_zipfile(self):
        z = zipfile.ZipFile(K64F_PACK_PATH, 'r')
        p = cmsis_pack.CmsisPack(z)
        pns = [x.part_number for x in p.devices]
        assert "MK64FN1M0xxx12" in pns

    def test_parse_device_info(self, k64f1m0):
        assert k64f1m0.vendor == "NXP"
        assert k64f1m0.families == ["MK64F12"]

    def test_get_svd(self, k64f1m0):
        svd = k64f1m0.svd
        x = ElementTree.parse(svd)
        assert x.getroot().tag == 'device'

    def test_mem_map(self, k64f1m0):
        map = k64f1m0.memory_map
        bm = map.get_boot_memory()
        assert bm.start == 0 and bm.length == 1 * 1024 * 1024
        ram = map.get_default_region_of_type(memory_map.MemoryType.RAM)
        assert ram.start == 0x20000000 and ram.length == 0x30000

    # Verify the flash region was converted correctly.
    # Note that the sector size will be 0 because CmsisPackDevice just prepares the flash region
    # for processing by FlmFlashRegionBuilder.
    def test_flash(self, k64f1m0):
        map = k64f1m0.memory_map
        flash = map.get_boot_memory()
        assert isinstance(flash, memory_map.FlashRegion)
        assert flash.start == 0 and flash.length == 1 * 1024 * 1024
        # assert flash.sector_size == 4096

class TestFLM:
    def test_algo(self, k64algo):
        i = k64algo.flash_info
#         print(i)
        assert i.type == 1
        assert i.start == 0
        assert i.size == 1 * 1024 * 1024
        assert i.page_size == 512
        assert i.sector_info_list == [(0, 4 * 1024)]

    def test_algo_dict_entry_points(self, k64algo):
        # Create the RAM region where we want the algo to be placed.
        ram = memory_map.RamRegion(0x20000000, length=0x10000)
        d = k64algo.get_pyocd_flash_algo(4096, ram)
        instr_len = len(d['instructions']) * 4
        load_addr = ram.end + 1 - instr_len
        assert d['load_address'] == load_addr
        assert d['pc_init'] == load_addr + 0x5
        assert d['pc_unInit'] == load_addr + 0x55
        assert d['pc_eraseAll'] == load_addr + 0x79
        assert d['pc_erase_sector'] == load_addr + 0xaf
        assert d['pc_program_page'] == load_addr + 0xc3

    def test_algo_dict_two_page_bufs(self, k64algo):
        # Create the RAM region where we want the algo to be placed.
        ram = memory_map.RamRegion(0x20000000, length=0x10000)
        d = k64algo.get_pyocd_flash_algo(k64algo.page_size, ram)
        instr_base = d['load_address']
        buf_top = align_down(instr_base, flash_algo.PackFlashAlgo._PAGE_BUFFER_ALIGN)
        buf1 = buf_top - k64algo.page_size
        buf2 = buf1 - k64algo.page_size
        assert d['page_buffers'] == [buf1, buf2]

    def test_algo_dict_one_page_buf(self, k64algo):
        # First get a full-sized algo allocation.
        ram = memory_map.RamRegion(0x20000000, length=0x10000)
        d = k64algo.get_pyocd_flash_algo(k64algo.page_size, ram)

        # Create a memory region with only enough memory for one page buf + stack.
        min_ram_size = len(d['instructions']) * 4 + k64algo.page_size + k64algo.page_size // 2
        min_ram = memory_map.RamRegion(0x20000000, length=min_ram_size)
        d = k64algo.get_pyocd_flash_algo(k64algo.page_size, min_ram)

        instr_base = d['load_address']
        assert instr_base == min_ram.end + 1 - len(d['instructions']) * 4
        buf_top = align_down(instr_base, flash_algo.PackFlashAlgo._PAGE_BUFFER_ALIGN)
        buf1 = buf_top - k64algo.page_size
        assert d['page_buffers'] == [buf1]

    # Flash Device:
    #   name=b'nRF53xxx_app'
    #   version=0x101
    #   type=1
    #   start=0x0
    #   size=0x200000
    #   page_size=0x1000
    #   value_empty=0xff
    #   prog_timeout_ms=1000
    #   erase_timeout_ms=3000
    #   sectors:
    #     start=0x0, size=0x1000
    def test_iter_sector_sizes_single(self, nrf5340appflm):
        sector_sizes = list(nrf5340appflm.iter_sector_size_ranges())
        assert sector_sizes == [
            (memory_map.MemoryRange(0x0000000, length=0x200000), 0x1000),
        ]

    # Flash Device:
    #   name=b'STM32F4xx 2MB Flash'
    #   version=0x101
    #   type=1
    #   start=0x8000000
    #   size=0x200000
    #   page_size=0x400
    #   value_empty=0xff
    #   prog_timeout_ms=100
    #   erase_timeout_ms=6000
    #   sectors:
    #     start=0x0, size=0x4000
    #     start=0x10000, size=0x10000
    #     start=0x20000, size=0x20000
    #     start=0x100000, size=0x4000
    #     start=0x110000, size=0x10000
    #     start=0x120000, size=0x20000
    def test_iter_sector_sizes_multiple(self, stm32f42mflm):
        sector_sizes = list(stm32f42mflm.iter_sector_size_ranges())
        assert sector_sizes == [
            (memory_map.MemoryRange(0x8000000, length=0x10000), 0x4000),
            (memory_map.MemoryRange(0x8010000, length=0x10000), 0x10000),
            (memory_map.MemoryRange(0x8020000, length=(0x100000 - 0x20000)), 0x20000),
            (memory_map.MemoryRange(0x8100000, length=0x10000), 0x4000),
            (memory_map.MemoryRange(0x8110000, length=0x10000), 0x10000),
            (memory_map.MemoryRange(0x8120000, length=(0x100000 - 0x20000)), 0x20000),
        ]

class TestFlmRegionBuilder:
    @pytest.fixture(scope='module')
    def builder(self):
        mock_target = MagicMock()
        mock_target.part_number = "TestPartNumber"
        ram = memory_map.RamRegion(0x20000000, length=0x10000, is_default=True)
        ram2 = memory_map.RamRegion(0x30010000, length=0x10000, is_default=False)
        memmap = memory_map.MemoryMap(ram, ram2)
        builder = FlmFlashRegionBuilder(mock_target, memmap)
        return builder

    def test_single_sector_size(self, builder: FlmFlashRegionBuilder, nrf5340appflm):
        flash = memory_map.FlashRegion(0, length=0x200000, flm=nrf5340appflm)
        assert builder.finalise_region(flash)
        assert not flash.has_subregions
        assert flash.sector_size == 0x1000

    def test_multiple_sector_size(self, builder: FlmFlashRegionBuilder, stm32f42mflm):
        flash = memory_map.FlashRegion(0x08000000, length=0x200000, flm=stm32f42mflm)
        assert builder.finalise_region(flash)
        assert flash.has_subregions
        submap = flash.submap
        assert submap.region_count == 6
        assert submap[0].sector_size == 0x4000
        assert submap[1].sector_size == 0x10000
        assert submap[2].sector_size == 0x20000
        assert submap[3].sector_size == 0x4000
        assert submap[4].sector_size == 0x10000
        assert submap[5].sector_size == 0x20000

    def test_ram_select_default(self, builder: FlmFlashRegionBuilder, nrf5340appflm):
        flash = memory_map.FlashRegion(0, length=0x200000, flm=nrf5340appflm)
        builder.finalise_region(flash)
        assert flash.algo
        instr_len = len(flash.algo['instructions']) * 4
        assert flash.algo['load_address'] == (0x20010000 - instr_len)
        assert not flash.has_subregions
        assert flash.algo

    def test_ram_select_explicit(self, builder: FlmFlashRegionBuilder, nrf5340appflm):
        flash = memory_map.FlashRegion(0, length=0x200000, flm=nrf5340appflm,
                                        _RAMstart=0x30010000, _RAMsize=0x4000)
        assert builder.finalise_region(flash)
        assert flash.algo
        instr_len = len(flash.algo['instructions']) * 4
        assert flash.algo['load_address'] == (0x30014000 - instr_len)
        assert not flash.has_subregions
        assert flash.algo

def has_overlapping_regions(memmap):
    return any((len(memmap.get_intersecting_regions(r.start, r.end)) > 1) for r in memmap.regions)

class TestNRF:
    def test_regions(self, nrf5340):
        memmap = nrf5340.memory_map
        assert not has_overlapping_regions(memmap)

class TestSTM32L4:
    def test_regions(self, stm32l4r5):
        memmap = stm32l4r5.memory_map
        assert not has_overlapping_regions(memmap)

class TestLPC55S36:
    def test_regions(self, lpc55s36):
        import pprint
        memmap = lpc55s36.memory_map
        print("memory map:")
        pprint.pprint(memmap.regions)
        assert not has_overlapping_regions(memmap)

class TestAlgoOverlappingAliasRegion:
    def test_regions(self, test2dev):
        import pprint
        memmap = test2dev.memory_map
        print("memory map:")
        pprint.pprint(memmap.regions)
        assert not has_overlapping_regions(memmap)
        # assert False

class TestAPID:
    def test1_dp(self, test1dev):
        assert test1dev.valid_dps == [0]

    def test1_procs(self, test1dev):
        procs = test1dev.processors_map
        m4 = procs['CM4']
        assert m4.name == 'CM4'
        assert m4.ap_address == APv1Address(0)
        assert m4.svd_path == "cm4.svd"
        m0p = procs['CM0p']
        assert m0p.name == 'CM0p'
        assert m0p.ap_address == APv1Address(2)
        assert m0p.svd_path == "cm0p.svd"




# pyOCD debugger
# Copyright (c) 2006-2020 Arm Limited
# Copyright (c) 2022-2023 Chris Reed
# Copyright (c) 2023 Hardy Griech
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pathlib import Path
import pytest
import os
import logging
# import telnetlib
import six

from pyocd.core.helpers import ConnectHelper
from pyocd.core.target import Target
from pyocd.debug import semihost
from pyocd.utility.server import StreamServer
from pyocd.utility.timeout import Timeout

@pytest.fixture(scope='module')
def tgt(request):
    session = None
    try:
        session = ConnectHelper.session_with_chosen_probe(blocking=False, return_first=True)
    except Exception as error:
        logging.error("Exception during session_with_chosen_probe", exc_info=error)
    if session is None:
        pytest.skip("No probe present")
        return
    session.open()
    session.options['resume_on_disconnect'] = False
    assert session.target
    session.target.reset_and_halt()

    def close_session():
        session.close()

    request.addfinalizer(close_session)
    return session.target

@pytest.fixture(scope='module')
def ctx(tgt):
    return tgt.get_target_context()

@pytest.fixture(scope='function')
def semihostagent(ctx, request):
    io_handler = semihost.InternalSemihostIOHandler()
    agent = semihost.SemihostAgent(ctx, io_handler)
    def cleanup():
        agent.cleanup()
    request.addfinalizer(cleanup)
    return agent

@pytest.fixture(scope='module')
def ramrgn(tgt):
    map = tgt.get_memory_map()
    for rgn in map:
        if rgn.is_ram:
            return rgn
    pytest.skip("No RAM available to load test")

def run_til_halt(tgt, semihostagent):
    with Timeout(2.0) as t:
        logging.info("Resuming target")
        tgt.resume()

        while True:
            if not t.check():
                tgt.halt()
                return False
            if tgt.get_state() == Target.State.HALTED:
                logging.info("Target halted")
                didHandle = semihostagent.check_and_handle_semihost_request()
                if didHandle:
                    logging.info("Semihost request handled")
                else:
                    logging.info("Non-semihost break")
                return didHandle

NOP = 0x46c0
BKPT_00 = 0xbe00
BKPT_AB = 0xbeab

class RecordingSemihostIOHandler(semihost.SemihostIOHandler):
    """@brief Semihost IO handler that records output.

    This handler is only meant to be used for console I/O since it doesn't implement
    open() or close().
    """
    def __init__(self):
        self._out_data = {}
        self._in_data = {}

    def set_input_data(self, fd, data):
        self._in_data[fd] = data

    def get_output_data(self, fd):
        if fd in self._out_data:
            return self._out_data[fd]
        else:
            return None

    def write(self, fd, ptr, length):
        assert self.agent
        if fd not in self._out_data:
            self._out_data[fd] = b''
        assert self.agent
        s = self.agent.get_data(ptr, length)
        self._out_data[fd] += s
        return 0

    def read(self, fd, ptr, length):
        assert self.agent
        if fd not in self._in_data:
            return length
        d = self._in_data[fd][:length]
        self._in_data[fd] = self._in_data[fd][length:]
        assert self.agent
        self.agent.context.write_memory_block8(ptr, bytearray(six.ensure_binary(d)))
        return length - len(d)

    def readc(self):
        if semihost.STDIN_FD not in self._in_data:
            return -1
        d = self._in_data[semihost.STDIN_FD][:1]
        self._in_data[semihost.STDIN_FD] = self._in_data[semihost.STDIN_FD][1:]
        if len(d):
            return ord(d[0])
        else:
            return -1

class SemihostRequestBuilder:
    """@brief Utility to build code and set registers to perform a semihost request."""
    def __init__(self, tgt, semihostagent, ramrgn):
        self.tgt = tgt
        self.ctx = tgt.get_target_context()
        self.semihostagent = semihostagent
        self.ramrgn = ramrgn

    def set_agent(self, agent):
        self.semihostagent = agent

    def setup_semihost_request(self, rqnum):
        assert self.tgt.get_state() == Target.State.HALTED

        self.ctx.write16(self.ramrgn.start, NOP)
        self.ctx.write16(self.ramrgn.start + 2, BKPT_AB)
        self.ctx.write16(self.ramrgn.start + 4, BKPT_00)

        self.ctx.write_core_register('pc', self.ramrgn.start)
        self.ctx.write_core_register('sp', self.ramrgn.start + 0x100)
        self.ctx.write_core_register('r0', rqnum)
        self.ctx.write_core_register('r1', self.ramrgn.start + 0x200)
        self.ctx.flush()
        return self.ramrgn.start + 0x200

    def do_open(self, filename, mode):
        argsptr = self.setup_semihost_request(semihost.SemihostingRequests.SYS_OPEN)

        # Write filename
        filename = bytearray(six.ensure_binary(filename) + b'\x00')
        self.ctx.write_memory_block8(argsptr + 12, filename)

        self.ctx.write32(argsptr, argsptr + 12) # null terminated filename
        self.ctx.write32(argsptr + 4, semihost.SemihostAgent.OPEN_MODES.index(mode)) # mode
        self.ctx.write32(argsptr + 8, len(filename) - 1) # filename length minus null terminator

        was_semihost = run_til_halt(self.tgt, self.semihostagent)
        assert was_semihost

        result = self.ctx.read_core_register('r0')
        return result

    def do_close(self, fd):
        argsptr = self.setup_semihost_request(semihost.SemihostingRequests.SYS_CLOSE)
        self.ctx.write32(argsptr, fd)

        was_semihost = run_til_halt(self.tgt, self.semihostagent)
        assert was_semihost

        result = self.ctx.read_core_register('r0')
        return result

    def do_write(self, fd, data):
        argsptr = self.setup_semihost_request(semihost.SemihostingRequests.SYS_WRITE)

        # Write data
        data = six.ensure_binary(data)
        self.ctx.write_memory_block8(argsptr + 12, data)

        self.ctx.write32(argsptr, fd) # fd
        self.ctx.write32(argsptr + 4, argsptr + 12) # data
        self.ctx.write32(argsptr + 8, len(data)) # data length
        self.ctx.flush()

        was_semihost = run_til_halt(self.tgt, self.semihostagent)
        assert was_semihost

        result = self.ctx.read_core_register('r0')
        return result

    def do_writec(self, c):
        argsptr = self.setup_semihost_request(semihost.SemihostingRequests.SYS_WRITEC)
        self.ctx.write8(argsptr, ord(c))

        was_semihost = run_til_halt(self.tgt, self.semihostagent)
        assert was_semihost

        result = self.ctx.read_core_register('r0')
        return result

    def do_write0(self, data):
        argsptr = self.setup_semihost_request(semihost.SemihostingRequests.SYS_WRITE0)

        data = data + b'\x00'
        self.ctx.write_memory_block8(argsptr, data)

        was_semihost = run_til_halt(self.tgt, self.semihostagent)
        assert was_semihost

        result = self.ctx.read_core_register('r0')
        return result

    def do_read(self, fd, length):
        argsptr = self.setup_semihost_request(semihost.SemihostingRequests.SYS_READ)

        # Clear read buffer.
        self.ctx.write_memory_block8(argsptr + 12, bytearray(b'\x00') * length)

        self.ctx.write32(argsptr, fd) # fd
        self.ctx.write32(argsptr + 4, argsptr + 12) # ptr
        self.ctx.write32(argsptr + 8, length) # data length
        self.ctx.flush()

        was_semihost = run_til_halt(self.tgt, self.semihostagent)
        assert was_semihost

        result = self.ctx.read_core_register('r0')

        # Read data put into read buffer.
        data = bytes(self.tgt.read_memory_block8(argsptr + 12, length - result))

        return result, data

    def do_seek(self, fd, pos):
        argsptr = self.setup_semihost_request(semihost.SemihostingRequests.SYS_SEEK)

        self.ctx.write32(argsptr, fd) # fd
        self.ctx.write32(argsptr + 4, pos) # pos
        self.ctx.flush()

        was_semihost = run_til_halt(self.tgt, self.semihostagent)
        assert was_semihost

        result = self.ctx.read_core_register('r0')

        return result

    def do_flen(self, fd):
        argsptr = self.setup_semihost_request(semihost.SemihostingRequests.SYS_FLEN)

        self.ctx.write32(argsptr, fd) # fd
        self.ctx.flush()

        was_semihost = run_til_halt(self.tgt, self.semihostagent)
        assert was_semihost

        result = self.ctx.read_core_register('r0')

        return result

    def do_istty(self, fd):
        argsptr = self.setup_semihost_request(semihost.SemihostingRequests.SYS_ISTTY)

        self.ctx.write32(argsptr, fd) # fd
        self.ctx.flush()

        was_semihost = run_til_halt(self.tgt, self.semihostagent)
        assert was_semihost

        result = self.ctx.read_core_register('r0')

        return result

    def do_no_args_call(self, rq):
        argsptr = self.setup_semihost_request(rq)
        self.ctx.write_core_register('r1', 0) # r1 must be zero on entry

        was_semihost = run_til_halt(self.tgt, self.semihostagent)
        assert was_semihost

        result = self.ctx.read_core_register('r0')
        return result

@pytest.fixture(scope='function')
def semihost_builder(tgt, semihostagent, ramrgn):
    return SemihostRequestBuilder(tgt, semihostagent, ramrgn)

@pytest.fixture(scope='function')
def console_semihost_builder(semihost_builder):
    console = RecordingSemihostIOHandler()
    agent = semihost.SemihostAgent(semihost_builder.ctx, console=console)
    semihost_builder.set_agent(agent)
    return semihost_builder

@pytest.fixture
def delete_testfile(request):
    def delete_it():
        try:
            os.remove("testfile")
        except IOError:
            pass
    request.addfinalizer(delete_it)

class TestSemihosting:
    """@brief Tests for semihost requests."""
    def test_open_stdio(self, semihost_builder):
        fd = semihost_builder.do_open(":tt", 'r') # stdin
        assert fd == 0

        fd = semihost_builder.do_open(":tt", 'w') # stdout
        assert fd == 1

        fd = semihost_builder.do_open(":tt", 'a') # stderr
        assert fd == 2

    def test_open_home_file(self, semihost_builder, request):
        testfilepath = Path("~/testfile").expanduser()

        def delete_it():
            try:
                testfilepath.unlink()
            except OSError:
                pass
        request.addfinalizer(delete_it)

        fd = semihost_builder.do_open("~/testfile", 'wb')
        assert fd > 2

        result = semihost_builder.do_write(fd, b"foo")
        assert result == 0

        result = semihost_builder.do_close(fd)
        assert result == 0

        data = testfilepath.read_bytes()
        assert data == b"foo"

    def test_open_close_file(self, semihost_builder, delete_testfile):
        fd = semihost_builder.do_open("testfile", 'w+b')
        assert fd > 2

        result = semihost_builder.do_close(fd)
        assert result == 0

    @pytest.mark.parametrize(("mode", "writeData", "pos", "readLen", "readResult"), [
            ("w+b", b"12345678", 0, 8, 0),                                      # several testcases handling binary data
            ("w+b", b"hi", 0, 2, 0),
            ("w+b", b"hello", 2, 3, 0),
            ("w+b", b"", 0, 4, 4),
            ("w+b", b"abcd", -1, 0, 0),
            ("w+", "", 0, 0, 0),                                                # write strings
            ("w+", "1", 0, 1, 0),
            ("w+", "12", 0, 2, 0),
            ("w+", "123", 0, 3, 0),
            ("w+", "1234", 0, 4, 0),
            ("w+", "Hello this is an extraaaaaaaa long string", 0, 41, 0),      # write a long string
            ("w+", "äöüÄÖÜ😀\u4500", 0, 6*2+4+3, 0),                           # write string with some UTF-8 encodings
        ])
    def test_file_write_read(self, semihost_builder, delete_testfile, mode, writeData, pos, readLen, readResult):
        fd = semihost_builder.do_open("testfile", mode)
        assert fd > 2

        if len(writeData):
            result = semihost_builder.do_write(fd, writeData)
            assert result == 0

            result = semihost_builder.do_flen(fd)
            assert result == len(six.ensure_binary(writeData))

        if pos != -1:
            result = semihost_builder.do_seek(fd, pos)
            assert result == 0

        result, data = semihost_builder.do_read(fd, readLen)
        assert result == readResult
        assert data == six.ensure_binary(writeData[pos:pos + readLen])

        result = semihost_builder.do_close(fd)
        assert result == 0

    def test_console_write(self, semihost_builder):
        console = RecordingSemihostIOHandler()
        agent = semihost.SemihostAgent(semihost_builder.ctx, console=console)
        semihost_builder.set_agent(agent)

        result = semihost_builder.do_write(semihost.STDOUT_FD, b'hello world')
        assert result == 0

        assert console.get_output_data(semihost.STDOUT_FD) == b'hello world'

    def test_console_write_binary_pattern(self, semihost_builder):
        console = RecordingSemihostIOHandler()
        agent = semihost.SemihostAgent(semihost_builder.ctx, console=console)
        semihost_builder.set_agent(agent)

        binary_pattern = bytes.fromhex('8152666f72706cff08000000000000000000000000000000020000000000000000000000000000000300000000000000000000000000000016000000')
        assert len(binary_pattern) == 60
        result = semihost_builder.do_write(semihost.STDOUT_FD, binary_pattern)
        assert result == 0
        assert console.get_output_data(semihost.STDOUT_FD) == binary_pattern

    def test_console_writec(self, semihost_builder):
        console = RecordingSemihostIOHandler()
        agent = semihost.SemihostAgent(semihost_builder.ctx, console=console)
        semihost_builder.set_agent(agent)

        for c in 'abcdef':
            result = semihost_builder.do_writec(c)
            assert result == 0

        assert console.get_output_data(semihost.STDOUT_FD) == b'abcdef'

    def test_console_write0(self, semihost_builder):
        console = RecordingSemihostIOHandler()
        agent = semihost.SemihostAgent(semihost_builder.ctx, console=console)
        semihost_builder.set_agent(agent)

        result = semihost_builder.do_write0(b'this is a very looooooooooooooooooooooooooooooooooooooooooooong string with more than 32 characters')
        assert result == 0

        assert console.get_output_data(semihost.STDOUT_FD) == b'this is a very looooooooooooooooooooooooooooooooooooooooooooong string with more than 32 characters'

    @pytest.mark.parametrize(("data", "readlen"), [
            (b"12345678", 8),
            (b"hi", 2),
            (b"hello", 3),
            (b"", 4),
            (b"abcd", 0)
        ])
    def test_console_read(self, semihost_builder, data, readlen):
        console = RecordingSemihostIOHandler()
        agent = semihost.SemihostAgent(semihost_builder.ctx, console=console)
        semihost_builder.set_agent(agent)

        console.set_input_data(semihost.STDIN_FD, data)

        result, resultData = semihost_builder.do_read(semihost.STDIN_FD, readlen)

        assert result == (readlen - min(readlen, len(data)))

        expectedData = data[:min(len(data), readlen)]
        assert resultData == expectedData

    def test_console_readc(self, semihost_builder):
        console = RecordingSemihostIOHandler()
        agent = semihost.SemihostAgent(semihost_builder.ctx, console=console)
        semihost_builder.set_agent(agent)

        console.set_input_data(semihost.STDIN_FD, 'x')

        result = semihost_builder.do_no_args_call(semihost.SemihostingRequests.SYS_READC)
        assert chr(result) == 'x'

    def test_clock(self, semihost_builder):
        result = semihost_builder.do_no_args_call(semihost.SemihostingRequests.SYS_CLOCK)
        assert result != -1
        assert result != 0
        logging.info("clock = %d cs", result)

        result2 = semihost_builder.do_no_args_call(semihost.SemihostingRequests.SYS_CLOCK)
        assert result2 != -1
        assert result2 != 0
        assert result2 > result
        logging.info("clock = %d cs", result2)

    def test_time(self, semihost_builder):
        result = semihost_builder.do_no_args_call(semihost.SemihostingRequests.SYS_TIME)
        assert result != 0
        logging.info("time = %d sec", result)

    def test_errno_no_err(self, semihost_builder):
        result = semihost_builder.do_no_args_call(semihost.SemihostingRequests.SYS_ERRNO)
        assert result == 0

    @pytest.mark.parametrize(("fd"), [
            (semihost.STDIN_FD),
            (semihost.STDOUT_FD),
            (semihost.STDERR_FD),
        ])
    def test_istty_stdio(self, semihost_builder, fd):
        result = semihost_builder.do_istty(fd)
        assert result == 1

    def test_istty_non_stdio(self, semihost_builder, delete_testfile):
        fd = semihost_builder.do_open("testfile", 'w+b')
        assert fd > 2

        result = semihost_builder.do_istty(fd)
        assert result == 0

        result = semihost_builder.do_close(fd)
        assert result == 0

    def test_feature_bits(self, semihost_builder):
        fd = semihost_builder.do_open(":semihosting-features", 'rb')
        assert fd > 2

        # Required requests: SYS_FLEN, SYS_ISTTY, SYS_SEEK, SYS_READ, SYS_CLOSE

        n = semihost_builder.do_flen(fd)
        assert n >= 5 # 4 magic bytes + at least 1 feature flags byte

        result = semihost_builder.do_istty(fd)
        assert result == 0

        result = semihost_builder.do_seek(fd, 0)
        assert result == 0

        # Verify header
        result, data = semihost_builder.do_read(fd, 4)
        assert result == 0
        assert data == b'SHFB'

        result = semihost_builder.do_seek(fd, 4)
        assert result == 0

        # Verify first feature byte.
        result, data = semihost_builder.do_read(fd, 1)
        assert result == 0
        assert (data[0] & 0x3) == 0x02 # Only check feature bits defined at the time this test was written.

        # Seek back to earlier.
        result = semihost_builder.do_seek(fd, 2)
        assert result == 0

        result, data = semihost_builder.do_read(fd, 2)
        assert result == 0
        assert data == b'FB'

        # Close.
        result = semihost_builder.do_close(fd)
        assert result == 0


@pytest.fixture(scope='function')
def telnet_server(request):
    telnet_server = StreamServer(
            0,          # port 0 to automatically allocate a free port
            True,       # local only
            "Semihost", # name
            False,      # is read only
            extra_info="test"
            )
    def stopit():
        telnet_server.stop()
    request.addfinalizer(stopit)
    return telnet_server

@pytest.fixture(scope='function')
def semihost_telnet_agent(ctx, telnet_server, request):
    semihost_console = semihost.ConsoleIOHandler(telnet_server)
    agent = semihost.SemihostAgent(ctx, console=semihost_console)
    def cleanup():
        agent.cleanup()
    request.addfinalizer(cleanup)
    return agent

@pytest.fixture(scope='function')
def semihost_telnet_builder(tgt, semihost_telnet_agent, ramrgn):
    return SemihostRequestBuilder(tgt, semihost_telnet_agent, ramrgn)

# Telnet based tests are commented out for the time being, until they can be rewritten
# to not use the telnetlib package that is now deprecated and will be removed in Python 3.13.

# @pytest.fixture(scope='function')
# def telnet_conn(request, telnet_server):
#     from time import sleep
#     # Sleep for a bit to ensure the semihost telnet server has started up in its own thread.
#     while not telnet_server.is_running:
#         sleep(0.005)
#     telnet = telnetlib.Telnet('localhost', telnet_server.port, 10.0)
#     def cleanup():
#         telnet.close()
#     request.addfinalizer(cleanup)
#     return telnet

# class TestSemihostingTelnet:
#     def test_connect(self, semihost_telnet_builder, telnet_conn):
#         result = semihost_telnet_builder.do_no_args_call(semihost.SemihostingRequests.TARGET_SYS_ERRNO)
#         assert result == 0

#     def test_write(self, semihost_telnet_builder, telnet_conn):
#         result = semihost_telnet_builder.do_write(semihost.STDOUT_FD, b'hello world')
#         assert result == 0

#         index, _, text = telnet_conn.expect([b'hello world'])
#         assert index != -1
#         assert text == b'hello world'

#     def test_writec(self, semihost_telnet_builder, telnet_conn):
#         for c in (bytes([i]) for i in b'xyzzy'):
#             result = semihost_telnet_builder.do_writec(c)
#             assert result == 0

#             index, _, text = telnet_conn.expect([c])
#             assert index != -1
#             assert text == c

#     def test_write0(self, semihost_telnet_builder, telnet_conn):
#         result = semihost_telnet_builder.do_write0(b'hello world')
#         assert result == 0

#         index, _, text = telnet_conn.expect([b'hello world'])
#         assert index != -1
#         assert text == b'hello world'

#     def test_read(self, semihost_telnet_builder, telnet_conn):
#         telnet_conn.write(b'hello world')

#         result, data = semihost_telnet_builder.do_read(semihost.STDIN_FD, 11)
#         assert result == 0
#         assert data == b'hello world'

#     def test_readc(self, semihost_telnet_builder, telnet_conn):
#         telnet_conn.write(b'xyz')

#         for c in 'xyz':
#             rc = semihost_telnet_builder.do_no_args_call(semihost.SemihostingRequests.TARGET_SYS_READC)
#             assert chr(rc) == c

class TestSemihostAgent:
    def test_no_io_handler(self, ctx):
        a = semihost.SemihostAgent(ctx, io_handler=None, console=None)
        assert type(a.io_handler) is semihost.SemihostIOHandler
        assert type(a.console) is semihost.SemihostIOHandler
        assert a.console is a.io_handler

    def test_only_io_handler(self, ctx):
        c = RecordingSemihostIOHandler()
        a = semihost.SemihostAgent(ctx, io_handler=c, console=None)
        assert a.io_handler is c
        assert a.console is c

    def test_only_console(self, ctx):
        c = RecordingSemihostIOHandler()
        a = semihost.SemihostAgent(ctx, io_handler=None, console=c)
        assert type(a.io_handler) is semihost.SemihostIOHandler
        assert a.console is c

@pytest.fixture
def ioh(ctx):
    handler = semihost.SemihostIOHandler()
    agent = semihost.SemihostAgent(ctx, io_handler=handler)
    return handler, agent

class TestSemihostIOHandlerBase:
    @pytest.mark.parametrize(("filename", "mode", "expectedFd"), [
            (":tt", 'r', semihost.STDIN_FD),
            (":tt", 'w', semihost.STDOUT_FD),
            (":tt", 'a', semihost.STDERR_FD),
            (":tt", 'r+b', -1),
            ("somefile", 'r+b', None),
        ])
    def test_std_open(self, ctx, ramrgn, ioh, filename, mode, expectedFd):
        handler, agent = ioh
        ctx.write_memory_block8(ramrgn.start, bytearray(six.ensure_binary(filename) + b'\x00'))
        assert handler._std_open(ramrgn.start, len(filename), mode) == (expectedFd, filename)

    @pytest.mark.parametrize(("op", "args"), [
            ('open', (0, 0, 'r')),
            ('close', (1,)),
            ('write', (1, 0, 0)),
            ('read', (1, 0, 0)),
            ('readc', tuple()),
            ('istty', (1,)),
            ('seek', (1, 0)),
            ('flen', (1,)),
            ('remove', (0, 0)),
            ('rename', (0, 0, 0, 0)),
        ])
    def test_unimplemented(self, op, args):
        handler = semihost.SemihostIOHandler()
        with pytest.raises(NotImplementedError):
            handler.__getattribute__(op)(*args)





# pyOCD debugger
# Copyright (c) 2020 Arm Limited
# Copyright (c) 2021-2023 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import logging
import pytest
from unittest import mock
from lark.lexer import Token as LarkToken
from lark.tree import Tree as LarkTree

from pyocd.core import exceptions
from pyocd.debug.sequences.scope import Scope
from pyocd.debug.sequences.sequences import (
    DebugSequenceSemanticError,
    DebugSequenceExecutionContext,
    DebugSequence,
    Block,
    WhileControl,
    IfControl,
    Parser,
    SemanticChecker,
    _ConstantFolder,
)
from pyocd.core.session import Session
from pyocd.probe.debug_probe import DebugProbe

# Substitute for SequenceFunctionDelegate for testing.
class SequenceFunctionsDelegateForTesting:
    def valid_fn_no_args(self):
        pass

    def valid_fn_1_arg(self, a: int):
        pass

    def valid_fn_3_arg(self, a: int, b: int, c: int):
        pass

    def valid_fn_varg(self, a: int, *vargs):
        pass

    def fn_with_str_parm(self, foo: str):
        pass

    def sequence(self, name: str):
        pass

class SequenceDelegateForTesting:
    # This same root scope instance must be return from all .get_root_scope() calls since it is
    # checked for by id in some tests.
    _root = Scope(name="unit-test-root")

    @property
    def sequences(self):
        return set()

    @property
    def cmsis_pack_device(self):
        return mock.MagicMock()

    def get_root_scope(self, context) -> Scope:
        self._root.set('rootvar', 42)
        return self._root

    def has_sequence_with_name(self, name, pname=None):
        return name == "valid"

    def get_sequence_with_name(self, name, pname=None):
        print(f"SequenceDelegateForTesting.get_sequence_by_name: name={name} pname={pname}")
        if name == "valid":
            x = mock.Mock()
            print(f"returning {x}")
            return x
        else:
            return None

    def get_protocol(self):
        return 2

    def get_connection_type(self):
        return 1

    def get_traceout(self):
        return 0

    def get_sequence_functions(self) -> SequenceFunctionsDelegateForTesting:
        return SequenceFunctionsDelegateForTesting()


class MockProbe:
    def __init__(self):
        self.wire_protocol = DebugProbe.Protocol.SWD

@pytest.fixture(scope='function')
def session():
    s = Session(None)
    setattr(s, '_probe', MockProbe())
    return s

@pytest.fixture(scope='function')
def delegate():
    return SequenceDelegateForTesting()

@pytest.fixture(scope='function')
def scope():
    s = Scope(name="fixture")
    s.set("a", 0)
    s.set("b", 128)
    return s

@pytest.fixture(scope='function')
def context(session, delegate):
    c = DebugSequenceExecutionContext(session, delegate, pname=None)
    return c

@pytest.fixture(scope='function')
def block_context(context: DebugSequenceExecutionContext):
    # Return a context set up to be able to run a Block.
    #
    # First create a sequence node with no children. Use the private _create_scope() method to get
    # the new sequence scope with the delegate's root scope at parent. Then push the sequence node
    # and scope onto the context stack. We can't just execute the sequence to do all that because
    # it will pop the context before returning, leaving us without the proper context for running
    # a block.
    seq = DebugSequence('test_sequence')
    scope = seq._create_scope(context)
    # Must manually enter the push context manager.
    ctxmgr = context._push(seq, scope)
    return context

class TestDebugSequenceScope:
    def test_name(self):
        s = Scope(name='test')
        assert s.name == 'test'

    def test_get_1(self, scope):
        assert scope.get('a') == 0
        assert scope.get('b') == 128
        assert len(scope) == 2

    def test_get_invalid(self, scope):
        with pytest.raises(KeyError):
            scope.get('missing')
        with pytest.raises(KeyError):
            scope.get('')

    def test_set_1(self, scope):
        scope.set('a', 10)
        assert scope.get('a') == 10
        assert scope.get('b') == 128

    def test_set_ro(self, scope):
        scope.set('x', 32, readonly=True)
        assert scope.is_read_only('x')
        assert scope.get('x') == 32
        with pytest.raises(RuntimeError):
            scope.set('x', 50)

    def test_freeze(self, scope):
        scope.set('y', 16)
        scope.freeze()
        with pytest.raises(RuntimeError):
            scope.set('a', 1)
        assert len(scope) == 3
        assert scope.is_read_only('b')

    def test_parent_1(self, scope):
        sub = Scope(parent=scope, name='subscope')
        assert sub.parent is scope
        assert sub.get('a') == 0
        assert sub.get('b') == 128
        assert sub.is_defined('b')

    def test_parent_2(self, scope):
        sub = Scope(parent=scope, name='subscope')
        sub.set('x', 1)
        assert sub.parent is scope
        assert sub.get('a') == 0
        assert sub.get('x') == 1
        assert sub.is_defined('b')
        assert not scope.is_defined('x')

    def test_parent_3(self, scope):
        sub = Scope(parent=scope, name='subscope')
        sub.set('a', 1)
        assert sub.parent is scope
        assert sub.get('a') == 1
        assert scope.get('a') == 1 # Make sure it was set in parent scope.
        assert sub.is_defined('a')
        assert 'a' not in sub._variables
        assert scope.is_defined('a')

    def test_parent_4(self, scope):
        sub = Scope(parent=scope, name='subscope')
        scope.set('y', 12, readonly=True)
        with pytest.raises(RuntimeError):
            sub.set('y', 1)
        assert sub.get('y') == 12
        assert scope.get('y') == 12

    def test_copy(self, scope):
        other = Scope(name='other')
        other.copy_variables(from_scope=scope, variables=['a', 'b'])
        assert other.is_defined('a') \
            and other.is_defined('b')
        assert other.get('a') == scope.get('a')
        assert other.get('b') == scope.get('b')

    def test_copy_undefined(self, scope):
        other = Scope(name='other')
        other.copy_variables(from_scope=scope, variables=['a', 'missing'])
        assert other.is_defined('a') \
            and not other.is_defined('b') \
            and not other.is_defined('missing')
        assert other.get('a') == scope.get('a')


class TestDebugSequenceParser:
    def test_semicolons(self):
        # first with semicolon
        ast = Parser().parse("12;")
        print("ast=", ast)
        assert ast.children[0].children[0] == 12

        # now without semicolon
        ast = Parser().parse("12")
        print("ast=", ast)
        assert ast.children[0].children[0] == 12

    def test_fncall_no_args(self):
        ast = Parser().parse("myfunc();")
        print("ast=", ast)
        fncall = ast.children[0].children[0] # start.expr_stmt...
        assert fncall.data == 'fncall'
        assert len(fncall.children) == 1 # IDENT only
        assert fncall.children[0] == LarkToken('IDENT', 'myfunc')

    def test_fncall_3_arg(self):
        ast = Parser().parse("myfunc(1, 2, 3);")
        print("ast=", ast)
        fncall = ast.children[0].children[0] # start.expr_stmt...
        assert fncall.data == 'fncall'
        assert len(fncall.children) == 4 # IDENT, arg1, arg2, arg3
        assert fncall.children[0] == LarkToken('IDENT', 'myfunc')
        assert fncall.children[1] == 1
        assert fncall.children[2] == 2
        assert fncall.children[3] == 3

    def test_bad_input(self):
        with pytest.raises(exceptions.Error):
            Parser().parse("bad input ••• wooo")

    def test_unary_op_assign(self):
        # Statement from Infineon.PSoC6_DFP
        ast = Parser().parse("__Result = -1; // DAP is unavailable")
        print("ast=", ast)

    def test_assign_minus_negative(self):
        ast = Parser().parse("a = 1 - -1;")
        print("ast=", ast)
        e = ast.children[0].children[0].children[2] # start.expr_stmt.assign_expr.binary_expr
        assert e.children[0] == 1
        assert e.children[1] == LarkToken('MINUS', '-')
        assert e.children[2].data == 'unary_expr'
        assert e.children[2].children[0] == LarkToken('MINUS', '-')
        assert e.children[2].children[1] == 1

    def test_assign_minus_positive(self):
        ast = Parser().parse("a = 1 - +1;")
        print("ast=", ast)
        e = ast.children[0].children[0].children[2] # start.expr_stmt.assign_expr.binary_expr
        assert e.children[0] == 1
        assert e.children[1] == LarkToken('MINUS', '-')
        assert e.children[2].data == 'unary_expr'
        assert e.children[2].children[0] == LarkToken('PLUS', '+')
        assert e.children[2].children[1] == 1

    @pytest.mark.parametrize("literal", [
            "1000", "0x3e8", "0x3E8", "0b1111101000",
            "1000u", "1000U", "0x3e8u", "0x3e8U", "0b1111101000u", "0b1111101000U",
        ])
    def test_unsigned_int_lit(self, literal):
        ast = Parser().parse(f"{literal};")
        print("ast=", ast)
        assert ast.children[0].children[0] == 1000

    def test_simple_assign_expr(self):
        ast = Parser().parse("x = 10;")
        print("ast=", ast)
        e = ast.children[0].children[0] # start.expr_stmt.assign_expr
        assert e.children[2] == 10

    def test_multi_assign_expr(self):
        # ast= start
        # expr_stmt
        #     assign_expr
        #     x
        #     =
        #     binary_expr
        #         assign_expr
        #         y
        #         +=
        #         binary_expr
        #             1
        #             +
        #             1
        #         +
        #         2
        ast = Parser().parse("x = (y += (1 + 1)) + 2;")
        print("ast=", ast.pretty())
        e = ast.children[0].children[0] # start.expr_stmt.assign_expr
        assert e.data == LarkToken('RULE', 'assign_expr')
        assert e.children[0] == 'x'
        e = e.children[2].children[0]
        assert e.data == LarkToken('RULE', 'assign_expr')
        assert e.children[0] == 'y'
        assert e.children[1] == "+="

    def test_assign_expr_precedence_1(self):
        ast = Parser().parse("x = 1 ? 2 : 3;")
        print("ast=", ast.pretty())
        e = ast.children[0].children[0] # start.expr_stmt.assign_expr
        assert e.data == LarkToken('RULE', 'assign_expr')

    def test_assign_expr_precedence_2(self):
        ast = Parser().parse("1 ? x = 2 : 3;")
        print("ast=", ast.pretty())
        e = ast.children[0].children[0] # start.expr_stmt.assign_expr
        assert e.data == LarkToken('RULE', 'ternary_expr')
        assert e.children[1].data == LarkToken('RULE', 'assign_expr')

class TestDebugSequenceBlockExecute:
    def test_semicolons(self, block_context):
        block_context.current_scope.set("a", 0)

        # First with semicolon.
        s = Block("a == 0;")
        s.execute(block_context)
        assert block_context.current_scope.get("a") == 0

        # Now without semicolon.
        s = Block("a == 0")
        s.execute(block_context)
        assert block_context.current_scope.get("a") == 0

    def test_set_var(self, block_context):
        s = Block("__var x = 100;")
        s.execute(block_context)
        assert block_context.current_scope.get("x") == 100

    def test_var_no_expr(self, block_context):
        s = Block("__var x;")
        s.execute(block_context)
        assert block_context.current_scope.get("x") == 0

    def test_var_no_expr_separate_set(self, block_context):
        s = Block("__var x; x = 123;")
        s.execute(block_context)
        assert block_context.current_scope.get("x") == 123

    @pytest.mark.parametrize(("expr", "result"), [
            ("-1", 0xffffffffffffffff),
            ("-2", 0xfffffffffffffffe),
            ("!1", 0),
            ("!0", 1),
            ("+1", 1),
            ("~0xffff", 0xffffffffffff0000),
        ])
    def test_int_unary_ops(self, block_context, expr, result):
        s = Block("__var x = %s;" % expr)
        s.execute(block_context)
        assert block_context.current_scope.get("x") == result

    @pytest.mark.parametrize(("expr", "result"), [
            ("1 + 1", 2),
            ("2 - 1", 1),
            ("2 * 4", 8),
            ("4 / 2", 2),
            ("5 % 4", 1),
            ("1 << 12", 4096),
            ("0x80 >> 4", 0x8),
            ("0b1000 | 0x2", 0b1010),
            ("0b1100 & 0b0100", 0b0100),
        ])
    def test_int_expr(self, block_context, expr, result):
        s = Block("__var x = %s;" % expr)
        s.execute(block_context)
        assert block_context.current_scope.get("x") == result

    @pytest.mark.parametrize(("expr", "result"), [
            ("1 == 1", 1),
            ("1 == 0", 0),
            ("0 == 1", 0),
            ("1 != 1", 0),
            ("1 != 0", 1),
            ("0 != 1", 1),
            ("20 > 10", 1),
            ("20 > 20", 0),
            ("20 > 100", 0),
            ("5 >= 2", 1),
            ("5 >= 5", 1),
            ("5 >= 100", 0),
            ("10 < 20", 1),
            ("10 < 10", 0),
            ("10 < 4", 0),
            ("10 <= 20", 1),
            ("10 <= 10", 1),
            ("10 <= 5", 0),
        ])
    def test_bool_cmp_expr(self, block_context, expr, result):
        s = Block("__var x = %s;" % expr)
        s.execute(block_context)
        assert block_context.current_scope.get("x") == result

    # Aside from the obvious, verify that && and || are evaluated as in C rather than Python.
    # That is, they must produce a 1 or 0 and not the value of either operand.
    @pytest.mark.parametrize(("expr", "result"), [
            ("1 && 1", 1),
            ("1 && 0", 0),
            ("0 && 1", 0),
            ("0 && 0", 0),
            ("1 || 1", 1),
            ("1 || 0", 1),
            ("0 || 1", 1),
            ("0 || 0", 0),
            ("5 && 1000", 1),
            ("432 && 0", 0),
            ("0 && 2", 0),
            ("0 && 0", 0),
            ("348 || 4536", 1),
            ("5 || 0", 1),
            ("0 || 199", 1),
            ("0 || 0", 0),
        ])
    def test_bool_and_or_expr(self, block_context, expr, result):
        s = Block("__var x = %s;" % expr)
        s.execute(block_context)
        assert block_context.current_scope.get("x") == result

    @pytest.mark.parametrize(("expr", "result"), [
            ("1 + 2 * 5", 11),
            ("7 * 12 + 5", 89),
            ("1 + 5 - 3", 3),
            ("(1 + 2) * 5", 15),
            ("1 + (2 * 5)", 11),
            ("2 + 16 / 2", 10),
            ("1 + 17 % 3", 3),
            ("2 * 3 * 4", 24),
            ("0 || 1 && 1", 1),
            ("0 && 1 || 0", 0),
            ("1 == 6 > 5", 1),
            ("1 != 6 < 12", 0),
            ("1 << 4 > 1 << 2", 1),
            ("1 << (4 > 1) << 2", 8),
            ("!1 == 0", 1),
        ])
    def test_precedence(self, block_context, expr, result):
        s = Block("__var x = %s;" % expr)
        logging.info("Block: %s", s._ast.pretty())
        s.execute(block_context)
        actual = block_context.current_scope.get("x")
        assert actual == result

    @pytest.mark.parametrize(("expr", "result"), [
            ("(7 * (1 << 3) + 1) >> 1", 28),
        ])
    def test_longer_expr(self, block_context, expr, result):
        s = Block("__var x = %s;" % expr)
        logging.info("Block: %s", s._ast.pretty())
        s.execute(block_context)
        actual = block_context.current_scope.get("x")
        assert actual == result

    def test_unary_op_assign_block(self, block_context):
        # Statement from Infineon.PSoC6_DFP
        s = Block("__Result = -1; // DAP is unavailable")
        logging.info(f"Block: {s._ast.pretty()}")
        s.execute(block_context)
        actual = block_context.current_scope.get("__Result")
        assert actual == 0xffffffffffffffff

    @pytest.mark.parametrize(("expr", "result"), [
            ("1 ? (1 + 1) : (1 - 1)", 2),
            ("0 ? 10: 20", 20),
            ("1 << 5 ? 17 * (2 + 1) : 0", 51),
        ])
    def test_ternary_expr(self, block_context, expr, result):
        s = Block(f"__var x = {expr};")
        logging.info("Block: %s", s._ast.pretty())
        s.execute(block_context)
        actual = block_context.current_scope.get("x")
        assert actual == result

    @pytest.mark.parametrize(("expr", "result"), [
            ("x += 1", 2),
            ("x -= 1", 0),
            ("x *= 10", 10),
            ("x /= 1", 1),
            ("x %= 1", 0),
            ("x &= 3", 1),
            ("x |= 0x40", 0x41),
            ("x ^= 3", 2),
            ("x <<= 5", 1 << 5),
            ("x >>= 0", 1),
        ])
    def test_compound_assign(self, block_context, expr, result):
        s = Block("__var x = 1; %s;" % expr)
        s.execute(block_context)
        assert block_context.current_scope.get("x") == result

    def test_assign_expr_block(self, block_context):
        s = Block("__var x; __var y; x = 1 + (y = 2);")
        logging.info(f"Block: {s._ast.pretty()}")
        s.execute(block_context)
        assert block_context.current_scope.get("x") == 3
        assert block_context.current_scope.get("y") == 2

class TestConstantFolder:
    def _get_folded_ast(self, expr):
        ast = Parser.parse(expr)
        logging.info("Unoptimized AST:\n%s", ast.pretty())
        ast_opt = _ConstantFolder().transform(ast)
        logging.info("Optimized AST:\n%s", ast_opt.pretty())
        return ast_opt

    def _do_fold_test(self, expr, expected):
        ast_opt = self._get_folded_ast(expr)
        assert isinstance(ast_opt, LarkTree)
        assert ast_opt.data == 'start'
        assert ast_opt.children[0].data == 'expr_stmt'
        assert ast_opt.children[0].children[0] == expected

    @pytest.mark.parametrize(("op", "expected"), [
            # return left
            ("+",   "x"),
            ("-",   "x"),
            ("|",   "x"),
            ("^",   "x"),
            ("<<",  "x"),
            (">>",  "x"),
            ("||",  "x"),
            # return 0
            ("*",   0),
            ("/",   0),
            ("%",   0),
            ("&",   0),
            ("&&",  0),
        ])
    def test_fold_left_0(self, op, expected):
        self._do_fold_test(f"x {op} 0", expected)

    @pytest.mark.parametrize(("op", "expected"), [
            # return right
            ("+",   "x"),
            ("-",   "x"),
            ("|",   "x"),
            ("^",   "x"),
            ("||",  "x"),
            # return 0
            ("*",   0),
            ("/",   0),
            ("%",   0),
            ("&",   0),
            ("<<",  0),
            (">>",  0),
            ("&&",  0),
        ])
    def test_fold_right_0(self, op, expected):
        self._do_fold_test(f"0 {op} x", expected)

    @pytest.mark.parametrize(("op", "expected"), [
            # return left
            ("*",   "x"),
            ("/",   "x"),
            # return 1
            ("||",  1),
            # return 0
            ("%",   0),
        ])
    def test_fold_left_1(self, op, expected):
        self._do_fold_test(f"x {op} 1", expected)

    # Really more of a parser test.
    def test_unary_parens(self):
        self._do_fold_test(f"! (1)", 0)

    # Really more of a parser test.
    def test_unary_unfolded(self):
        ast_opt = self._get_folded_ast("! x")

        assert isinstance(ast_opt, LarkTree)
        assert ast_opt.data == 'start'
        assert ast_opt.children[0].data == 'expr_stmt'
        assert ast_opt.children[0].children[0].data == 'unary_expr'

    @pytest.mark.parametrize(("expr", "expected"), [
            ("-1",   0xffffffffffffffff),
            ("-0",   0),
        ])
    def test_unary_fold(self, expr, expected):
        self._do_fold_test(expr, expected)

    @pytest.mark.parametrize(("expr", "expected"), [
            ("1 ? 10 : 20",     10),
            ("0 ? 1 : 2",       2),
            ("0 ? 1 : 2 + 2",   4),
            ("1 + 1 ? 99 : 77", 99),
        ])
    def test_ternary_fold(self, expr, expected):
        self._do_fold_test(expr, expected)

class TestSemanticChecker:

    @pytest.mark.parametrize("expr", [
            "funkymonkey();",                   # fncall: invalid function name
            "valid_fn_1_arg(2134, x << 4);",    # fncall: too many args
            "valid_fn_1_arg(\"a-string\");",    # fncall: string arg for int param
            "valid_fn_3_arg(1, 2);",            # fncall: too few args
            "fn_with_str_parm(123);",           # fncall: int arg for string param
            "valid_fn_varg();",                 # vararg fn: too few args
            '"just a string in this expr";',    # expr consisting of only a string
            "__var a = \"string value\";",      # decl: attempt to assign string to variable
            "b = \"string value\";",            # assign: attempt to assign string to variable
            "Sequence();",                      # fn-specific check: no args
            "Sequence(1000);",                  # fn-specific check: int arg for str arg
        ])
    def test_sem_checker_raises(self, expr, context, scope):
        ast = Parser.parse(expr)
        logging.info("ast:\n%s", ast.pretty())
        c = SemanticChecker(ast, scope, context)
        with pytest.raises(DebugSequenceSemanticError):
            c.check()

    @pytest.mark.parametrize("expr", [
            "valid_fn_no_args();",
            "valid_fn_1_arg(2134);",
            "valid_fn_3_arg(1, 2, 3);",
            "fn_with_str_parm(\"bubblegum\");", # fncall: string arg for string param
            "valid_fn_varg(a);",                # vararg fn: exact number of positional args
            "valid_fn_varg(123, x, q + 1, \"hi there\", 99);",  # vararg fn: several var args
            "__var a = 3 + x * 2;",             # decl: valid assignment
            "b = 3 + x * 2;",                   # assign: valid assignment
            "Sequence(\"valid\");",             # fn-specific check: valid call
        ])
    def test_sem_checker_passes(self, expr, context, scope):
        ast = Parser.parse(expr)
        logging.info("ast:\n%s", ast.pretty())
        c = SemanticChecker(ast, scope, context)
        c.check()

class TestDebugSequences:
    def test_pname(self):
        seq = DebugSequence('test', pname="cm4")
        assert seq.name == 'test'
        assert seq.pname == 'cm4'

        assert DebugSequence('test').pname is None

    def test_info(self):
        assert DebugSequence('test').info == ''
        assert DebugSequence('test', info='hi there').info == 'hi there'

    def test_enable(self):
        assert DebugSequence('test').is_enabled
        assert not DebugSequence('test', is_enabled=False).is_enabled

    def test_scope_create(self, context):
        seq = DebugSequence('test')
        scope = seq._create_scope(context)
        assert scope.parent is context.delegate.get_root_scope(context)
        assert scope.get('__Result') == 0
        assert not scope.is_read_only('_Result')
        assert scope.get('__dp') == 0
        assert not scope.is_read_only('__dp')
        assert scope.get('__ap') == 0
        assert not scope.is_read_only('__ap')
        assert scope.get('__apid') == 0
        assert not scope.is_read_only('__apid')
        assert scope.get('__errorcontrol') == 0
        assert not scope.is_read_only('__errorcontrol')
        assert scope.get('__protocol') == 2
        assert scope.is_read_only('__protocol')
        assert scope.get('__connection') == 1
        assert scope.is_read_only('__connection')
        assert scope.get('__traceout') == 0
        assert scope.is_read_only('__traceout')
        assert scope.get('__FlashOp') == 0
        assert scope.is_read_only('__FlashOp')
        assert scope.get('__FlashAddr') == 0
        assert scope.is_read_only('__FlashAddr')
        assert scope.get('__FlashLen') == 0
        assert scope.is_read_only('__FlashLen')
        assert scope.get('__FlashArg') == 0
        assert scope.is_read_only('__FlashArg')

    def test_scope_create_with_calling_seq(self, block_context):
        # Modify some of the stacked scope's special vars.
        super_scope = block_context.current_scope
        super_scope.set('__dp', 1)
        super_scope.set('__ap', 2)
        super_scope.set('__apid', 3)
        super_scope.set('__errorcontrol', 1)
        super_scope.set('__Result', 333)

        # Create a sub-sequence.
        seq = DebugSequence('test')
        # _create_scope() will see the stacked sequence in the context.
        scope = seq._create_scope(block_context)

        # Certain vars are propagated to subsequences.
        assert scope.get('__dp') == 1
        assert scope.get('__ap') == 2
        assert scope.get('__apid') == 3
        assert scope.get('__errorcontrol') == 1
        # __Result should not have been propagated.
        assert scope.get('__Result') == 0

        # Modify the propagated variables.
        scope.set('__dp', 2)
        scope.set('__ap', 1)
        scope.set('__apid', 5)
        scope.set('__errorcontrol', 0)
        scope.set('__Result', 555)

        # Verify the variables weren't set in the supersequences.
        assert super_scope.get('__dp') == 1
        assert super_scope.get('__ap') == 2
        assert super_scope.get('__apid') == 3
        assert super_scope.get('__errorcontrol') == 1
        assert super_scope.get('__Result') == 333

    def test_exec_block(self, context):
        seq = DebugSequence('test')
        seq.add_child(Block("__var x = 1 + 1;"))
        seq.execute(context)

    def test_exec_if(self, context):
        seq = DebugSequence('test')
        seq.add_child(Block("__var x = 1;"))
        seq.add_child(IfControl("x"))
        seq.execute(context)

    def test_exec_while(self, context):
        seq = DebugSequence('test')
        seq.add_child(Block("__var x = 0;"))
        w = WhileControl("x < 2")
        w.add_child(Block("x += 1;"))
        seq.add_child(w)
        seq.execute(context)

    def test_exec_while_with_assign(self, context: DebugSequenceExecutionContext):
        seq = DebugSequence('test')
        seq.add_child(Block("__var x = 0; __var y = 0;"))
        w = WhileControl("x < (y = 2)")
        w.add_child(Block("x += 1;"))
        seq.add_child(w)
        seq.execute(context)





# pyOCD debugger
# Copyright (c) 2023 Arm Limited
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional
from ...coresight.coresight_target import CoreSightTarget
from ...core.memory_map import (RamRegion, MemoryMap)

from ...core.target import Target

# see https://developer.arm.com/documentation/101104/0200/programmers-model/system-control-element/system-control-register-block
SYSTEM_CONTROL = 0x50021000
RESET_MASK = SYSTEM_CONTROL + 0x104
RESET_MASK_SYSRSTREQ0_EN = 1 << 4
RESET_MASK_SYSRSTREQ1_EN = 1 << 5
CPU_WAIT = SYSTEM_CONTROL + 0x118
CPU_WAIT_CPU0 = 1
CPU_WAIT_CPU1 = 2

class AN521(CoreSightTarget):

    VENDOR = "Arm"

    MEMORY_MAP = MemoryMap(
        RamRegion(  name='code_ns',     start=0x00000000, length=0x08000000, access='rwx'),
        RamRegion(  name='code_s',      start=0x10000000, length=0x08000000, access='rwxs'),

        RamRegion(  name='sram_ns',     start=0x20000000, length=0x02000000, access='rwx'),
        RamRegion(  name='mtb_ns',      start=0x24000000, length=0x00004000, access='rwx'),
        RamRegion(  name='sram2_ns',    start=0x28000000, length=0x00200000, access='rwx'),
        RamRegion(  name='sram3_ns',    start=0x28200000, length=0x00200000, access='rwx'),

        RamRegion(  name='sram_s',      start=0x30000000, length=0x02000000, access='rwxs'),
        RamRegion(  name='mtb_s',       start=0x34000000, length=0x00004000, access='rwxs'),
        RamRegion(  name='sram2_s',     start=0x38000000, length=0x00200000, access='rwxs'),
        RamRegion(  name='sram3_s',     start=0x38200000, length=0x00200000, access='rwxs'),
        # External Parallel SRAM only mapped to non-secure
        RamRegion(  name='psram_ns',    start=0x80000000, length=0x01000000, access='rwx'),
    )

    def __init__(self, session):
        super().__init__(session, self.MEMORY_MAP)

    def create_init_sequence(self):
        seq = super().create_init_sequence()

        seq.insert_before('halt_on_connect',
            ('enable_sysresetreq',        self._enable_sysresetreq),
        )

        return seq

    def _enable_sysresetreq(self):
        reset_mask = self.read32(RESET_MASK)
        reset_mask |= RESET_MASK_SYSRSTREQ0_EN
        self.write32(RESET_MASK, reset_mask)


    def reset_and_halt(self, reset_type: Optional[Target.ResetType] = None):
        self.write32(CPU_WAIT, CPU_WAIT_CPU1)
        super().reset_and_halt(reset_type)



# pyOCD debugger
# Copyright (c) 2023 Northern Mechatronics, Inc.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from ...coresight.cortex_m import CortexM

LOG = logging.getLogger(__name__)

class AMA3BFamily(CortexM):
    REG_MCU_CTRL_BOOTLOADER = 0x400201A0
    REG_MCU_CTRL_SCRATCH0 = 0x400201B0

    def set_reset_catch(self, reset_type=None):
        # If Debugger Support is disabled by the SDBG bit in INFO0_SECURITY,
        # The least significant bit of register REG_MCU_CTRL_SCRATCH0 must be
        # set to indicate that a halt is requested by the debugger after
        # primary boot.
        #
        # Refer to document A-SOCA3B-UGGA02EN for more details.
        
        # Check the REG_MCU_CTRL_BOOTLOADER register to see if secure boot
        # is enabled for:
        #   bit 31:30 warm reset
        #   bit 29:28 cold reset
        #   bit 27:26 secure boot feature enabled
        secure_boot = False
        reg_bootloader = self.read_memory(self.REG_MCU_CTRL_BOOTLOADER)
        if (reg_bootloader & 0xFC000000):
            secure_boot = True
        LOG.debug("AMA3B Secure Boot: %x" % secure_boot)
        
        if(secure_boot is True):
            # Modify only the least significant bit and preserve the scratch
            # register as it could be used by the application firmware.
            reg_scratch0 = self.read_memory(self.REG_MCU_CTRL_SCRATCH0) | 0x01
            self.write_memory(self.REG_MCU_CTRL_SCRATCH0, reg_scratch0)
        else:
            LOG.debug("normal_set_reset_catch")
            super().set_reset_catch(reset_type)



# pyOCD debugger
# Copyright (c) 2023 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pyocd.debug.svd.loader import (
    SVDFile,
    SVDLoader,
)

class TestIntervalSvdAccess:
    def builtin_svd(self, name: str) -> SVDLoader:
        def completion(dev):
            pass
        loader = SVDLoader(SVDFile.from_builtin(name), completion)
        loader.run()
        return loader

    def test_load(self):
        loader = self.builtin_svd('Musca_B1.svd')
        assert loader.device
        assert [p for p in loader.device.peripherals if p.name == 'UART0']




# pyOCD debugger
# Copyright (c) 2006-2013 Arm Limited
# Copyright (c) 2019 Monadnock Systems Ltd.
# Copyright (c) 2024 Nordic Semiconductor ASA
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from time import sleep
import os
import re
from zipfile import ZipFile
from tempfile import TemporaryDirectory
from intelhex import IntelHex

from ...core import exceptions
from ...core.memory_map import FlashRegion
from ...core.target import Target
from ...coresight.coresight_target import CoreSightTarget
from ...flash.eraser import FlashEraser
from ...flash.file_programmer import FileProgrammer
from ...utility.timeout import Timeout
from ...utility.progress import print_progress
from ...commands.base import CommandBase
from ...commands.execution_context import CommandSet

AHB_AP_NUM = 0x0
AUX_AHB_AP_NUM = 0x1
CTRL_AP_NUM = 0x2

CTRL_AP_RESET = 0x000
CTRL_AP_ERASEALL = 0x004
CTRL_AP_ERASEALLSTATUS = 0x008
CTRL_AP_ERASEPROTECTSTATUS = 0x00C
CTRL_AP_ERASEPROTECTDISABLE = 0x010
CTRL_AP_APPROTECTSTATUS = 0x014
CTRL_AP_MAILBOX_TXDATA = 0x020
CTRL_AP_MAILBOX_TXSTATUS = 0x024
CTRL_AP_MAILBOX_RXDATA = 0x028
CTRL_AP_MAILBOX_RXSTATUS = 0x02C
CTRL_AP_IDR = 0x0FC

CTRL_AP_ERASEALLSTATUS_READY = 0x0
CTRL_AP_ERASEALLSTATUS_READYTORESET = 0x1
CTRL_AP_ERASEALLSTATUS_BUSY = 0x2
CTRL_AP_ERASEALLSTATUS_ERROR = 0x3

CTRL_AP_APPROTECTSTATUS_APPROTECT_MSK = 0x1
CTRL_AP_APPROTECTSTATUS_SECUREAPPROTECT_MSK = 0x2
CTRL_AP_ERASEPROTECTSTATUS_MSK = 0x1

CTRL_IDR_EXPECTED = 0x32880000

MASS_ERASE_TIMEOUT = 30.0

CSW_DEVICEEN =  0x00000040

LOG = logging.getLogger(__name__)

def bytes_to_word(bts):
    result = 0
    for i, b in enumerate(bts):
        result |= b << (8*i)
    return result

def word_to_bytes(wrd):
    result = []
    for i in range(4):
        result.append((wrd >> (8*i)) & 0xFF)
    return bytes(result)

class NRF54L(CoreSightTarget):

    VENDOR = "Nordic Semiconductor"

    def __init__(self, session, memory_map=None):
        super(NRF54L, self).__init__(session, memory_map)
        self.ctrl_ap = None
        self.was_locked = False

    def create_init_sequence(self):
        seq = super(NRF54L, self).create_init_sequence()

        # Must check whether security is enabled, and potentially auto-unlock, before
        # any init tasks that require system bus access.
        seq.wrap_task('discovery',
            lambda seq: seq.insert_before('find_components',
                              ('check_ctrl_ap_idr', self.check_ctrl_ap_idr),
                              ('check_flash_security', self.check_flash_security),
                          )
            )
        seq.wrap_task('discovery',
            lambda seq: seq.insert_after('create_aps', ('fixup_rom_base', self.fixup_rom_base))
            )
        seq.insert_before('post_connect_hook',
                          ('check_part_info', self.check_part_info))

        return seq

    def fixup_rom_base(self):
        self.aps[0].rom_addr = 0xE00FE000
        self.aps[0].has_rom_table = True

    def check_ctrl_ap_idr(self):
        LOG.info("Checking CTRL-AP IDR")
        self.ctrl_ap = self.dp.aps[CTRL_AP_NUM]

        # Check CTRL-AP ID.
        if self.ctrl_ap.idr != CTRL_IDR_EXPECTED:
            LOG.error("%s: bad CTRL-AP IDR (is 0x%08x)", self.part_number, self.ctrl_ap.idr)

    def ap_is_enabled(self):
        csw = self.dp.read_ap(AHB_AP_NUM << 24)
        return csw & CSW_DEVICEEN

    def check_flash_security(self):
        """@brief Check security and unlock device.

        This init task determines whether the device is locked (APPROTECT enabled). If it is,
        and if auto unlock is enabled, then perform a mass erase to unlock the device.

        This init task runs *before* cores are created.
        """

        target_id = self.dp.read_dp(0x24)
        if target_id & 0xFFF != 0x289:
            LOG.error(f"This doesn't look like a Nordic Semiconductor device!")
        if target_id & 0xF0000 != 0xC0000:
            LOG.error(f"This doesn't look like an nRF54L device!")

        if not self.ap_is_enabled():
            if self.session.options.get('auto_unlock'):
                LOG.warning("%s APPROTECT enabled: will try to unlock via mass erase", self.part_number)

                self.mass_erase()
        else:
            LOG.warning("%s is not in a secure state", self.part_number)


    def check_part_info(self):
        partno = self.read32(0x00FFC31C)
        variant = self.read32(0x00FFC320)

        LOG.info(f"This appears to be an nRF{partno:X} " +
                 f"{word_to_bytes(variant).decode('ASCII', errors='ignore')}")

        deviceaddr = (self.read32(0x00FFC3A4),self.read32(0x00FFC3A8))
        mac_bytes = list(word_to_bytes(deviceaddr[0]) + word_to_bytes(deviceaddr[1])[:2])
        mac_bytes[5] |= 0xc0 # Set a Bluetooth LE random address as a static address
        mac = ":".join(f"{x:02X}" for x in mac_bytes[::-1])
        LOG.info(f"BLE MAC: {mac}")

    def mass_erase(self):
        # See Nordic Whitepaper nWP-027 for magic numbers and order of operations from the vendor
        self.ctrl_ap.write_reg(CTRL_AP_ERASEALL, 1)
        with Timeout(MASS_ERASE_TIMEOUT) as to:
            while to.check():
                status = self.ctrl_ap.read_reg(CTRL_AP_ERASEALLSTATUS)
                if status == CTRL_AP_ERASEALLSTATUS_BUSY:
                    break
                sleep(0.5)
            else:
                # Timed out
                LOG.error("Mass erase timeout waiting for ERASEALLSTATUS")
                return False
        with Timeout(MASS_ERASE_TIMEOUT) as to:
            while to.check():
                status = self.ctrl_ap.read_reg(CTRL_AP_ERASEALLSTATUS)
                if status == CTRL_AP_ERASEALLSTATUS_READYTORESET:
                    break
                sleep(0.5)
            else:
                # Timed out
                LOG.error("Mass erase timeout waiting for ERASEALLSTATUS")
                return False
        sleep(0.01)
        self.ctrl_ap.write_reg(CTRL_AP_RESET, 2)
        self.ctrl_ap.write_reg(CTRL_AP_RESET, 0)
        sleep(0.2)
        return True



#!/usr/bin/env python3
# pyOCD debugger
# Copyright (c) 2011-2021 Arm Limited
# Copyright (c) 2021-2022 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import argparse
import colorama
from datetime import datetime
import struct
import binascii
import jinja2
from pyocd.target.pack.flash_algo import PackFlashAlgo
from pyocd.utility.mask import align_up

# This header consists of two instructions:
#
# ```
# bkpt  #0
# b     .-2     # branch to the bkpt
# ```
#
# Before running a flash algo operation, LR is set to the address of the `bkpt` instruction,
# so when the operation function returns it will halt the CPU.
BLOB_HEADER = '0xe7fdbe00,'
HEADER_SIZE = 4

STACK_SIZE = 0x1000

PYOCD_TEMPLATE = \
"""# pyOCD debugger
# Copyright (c) {{year}} {{copyright_owner}}
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

FLASH_ALGO = {
    'load_address' : {{'0x%08x' % entry}},

    # Flash algorithm as a hex string
    'instructions': [
    {{prog_header}}
    {{algo.format_algo_data(4, 8, "c")}}
    ],

    # Relative function addresses
    'pc_init': {{'0x%08x' % (algo.symbols['Init'] + header_size + entry)}},
    'pc_unInit': {{'0x%08x' % (algo.symbols['UnInit'] + header_size + entry)}},
    'pc_program_page': {{'0x%08x' % (algo.symbols['ProgramPage'] + header_size + entry)}},
    'pc_erase_sector': {{'0x%08x' % (algo.symbols['EraseSector'] + header_size + entry)}},
    'pc_eraseAll': {{'0x%08x' % (algo.symbols['EraseChip'] + header_size + entry)}},

    'static_base' : {{'0x%08x' % entry}} + {{'0x%08x' % header_size}} + {{'0x%08x' % algo.rw_start}},
    'begin_stack' : {{'0x%08x' % stack_pointer}},
    'end_stack' : {{'0x%08x' % stack_base}},
    'begin_data' : {{'0x%08x' % entry}} + 0x1000,
    'page_size' : {{'0x%x' % algo.page_size}},
    'analyzer_supported' : False,
    'analyzer_address' : 0x00000000,
    # Enable double buffering
    'page_buffers' : [
        {{'0x%08x' % (page_buffers[0])}},
        {{'0x%08x' % (page_buffers[1])}}
    ],
    'min_program_length' : {{'0x%x' % algo.page_size}},

    # Relative region addresses and sizes
    'ro_start': {{'0x%x' % (header_size + algo.ro_start)}},
    'ro_size': {{'0x%x' % algo.ro_size}},
    'rw_start': {{'0x%x' % (header_size + algo.rw_start)}},
    'rw_size': {{'0x%x' % algo.rw_size}},
    'zi_start': {{'0x%x' % (header_size + algo.zi_start)}},
    'zi_size': {{'0x%x' % algo.zi_size}},

    # Flash information
    'flash_start': {{'0x%x' % algo.flash_start}},
    'flash_size': {{'0x%x' % algo.flash_size}},
    'sector_sizes': (
    {%- for start, size  in algo.sector_sizes %}
        {{ "(0x%x, 0x%x)" % (start, size) }},
    {%- endfor %}
    )
}

"""

colorama.init()

def str_to_num(val):
    return int(val, 0)  #convert string to number and automatically handle hex conversion

class PackFlashAlgoGenerator(PackFlashAlgo):
    """
    Class to wrap a flash algo

    This class is intended to provide easy access to the information
    provided by a flash algorithm, such as symbols and the flash
    algorithm itself.
    """

    def format_algo_data(self, spaces, group_size, fmt):
        """"
        Return a string representing algo_data suitable for use in a template

        The string is intended for use in a template.

        :param spaces: The number of leading spaces for each line
        :param group_size: number of elements per line (element type
            depends of format)
        :param fmt: - format to create - can be either "hex" or "c"
        """
        padding = " " * spaces
        if fmt == "hex":
            blob = binascii.b2a_hex(self.algo_data).decode()
            line_list = []
            for i in range(0, len(blob), group_size):
                line_list.append('"' + blob[i:i + group_size] + '"')
            return ("\n" + padding).join(line_list)
        elif fmt == "c":
            blob = self.algo_data[:]
            pad_size = 0 if len(blob) % 4 == 0 else 4 - len(blob) % 4
            blob = blob + b"\x00" * pad_size
            integer_list = struct.unpack("<" + "L" * (len(blob) // 4), blob)
            line_list = []
            for pos in range(0, len(integer_list), group_size):
                group = ["0x%08x" % value for value in
                         integer_list[pos:pos + group_size]]
                line_list.append(", ".join(group))
            return (",\n" + padding).join(line_list)
        else:
            raise ValueError("Unsupported format %s" % fmt)

    def process_template(self, template_text, data_dict=None):
        """
        Generate output from the supplied template

        All the public methods and fields of this class can be accessed from
        the template via "algo".

        :param template_path: Relative or absolute file path to the template
        :param data_dict: Additional data to use when generating
        """
        if data_dict is None:
            data_dict = {}
        else:
            assert isinstance(data_dict, dict)
            data_dict = dict(data_dict)
        assert "algo" not in data_dict, "algo already set by user data"
        data_dict["algo"] = self

        template = jinja2.Template(template_text)
        return template.render(data_dict)

def main():
    parser = argparse.ArgumentParser(description="Blob generator")
    parser.add_argument("elf_path", help="Elf, axf, or flm to extract flash algo from")
    parser.add_argument("--blob-start", default=0x20000000, type=str_to_num, help="Starting "
                        "address of the flash blob in target RAM. (default 0x20000000)")
    parser.add_argument("--stack-size", default=STACK_SIZE, type=str_to_num, help="Stack size for the algo. "
                        f"(default {STACK_SIZE})")
    parser.add_argument("-i", "--info-only", action="store_true", help="Only print information about the flash "
                        "algo, do not generate a blob.")
    parser.add_argument("-o", "--output", default="pyocd_blob.py", help="Path of output file "
                        "(default 'pyocd_blob.py').")
    parser.add_argument("-t", "--template", help="Path to Jinja template file (default is an internal "
                        "template for pyocd).")
    parser.add_argument('-c', '--copyright', help="Set copyright owner.")
    args = parser.parse_args()

    if not args.copyright and not args.info_only:
        print(f"{colorama.Fore.YELLOW}Warning! No copyright owner was specified. Defaulting to \"PyOCD Authors\". "
            f"Please set via --copyright, or edit output.{colorama.Style.RESET_ALL}")

    if args.template:
        with open(args.template, "r") as tmpl_file:
            tmpl = tmpl_file.read()
    else:
        tmpl = PYOCD_TEMPLATE

    with open(args.elf_path, "rb") as file_handle:
        algo = PackFlashAlgoGenerator(file_handle)

        print(algo.flash_info)

        # Page buffer base begins after algo and its rw/zi data, rounded up to 16 bytes.
        buffer_base = align_up(args.blob_start + HEADER_SIZE
                        + algo.ro_size + algo.rw_size + algo.zi_size, 0x10)

        page_buffers = [
            buffer_base,
            buffer_base + algo.page_size,
        ]

        # Allocate stack after buffers, with top and bottom rounded to 8 bytes.
        stack_base = align_up(buffer_base + algo.page_size * 2, 8)
        sp = align_up(stack_base + args.stack_size, 8)

        header_end = args.blob_start + HEADER_SIZE
        print(f"load addr:   {args.blob_start:#010x}")
        print(f"data:        {HEADER_SIZE + len(algo.algo_data):#x} bytes")
        print(f"  header:    {args.blob_start:#010x} + {HEADER_SIZE:#x} bytes")
        print(f"  ro:        {header_end + algo.ro_start:#010x} + {algo.ro_size:#x} bytes")
        print(f"  rw:        {header_end + algo.rw_start:#010x} + {algo.rw_size:#x} bytes")
        print(f"  zi:        {header_end + algo.zi_start:#010x} + {algo.zi_size:#x} bytes")
        print(f"buffer[0]:   {page_buffers[0]:#010x}")
        print(f"buffer[1]:   {page_buffers[1]:#010x}")
        print(f"stack:       {stack_base:#010x} .. {sp:#010x} ({sp - stack_base:#x} bytes)")

        print("\nSymbol offsets:")
        for n, v in algo.symbols.items():
            print(f"{n}:{' ' * (11 - len(n))} {v:#010x}")

        if args.info_only:
            return

        if len(algo.sector_sizes) > 1:
            print(f"{colorama.Fore.YELLOW}Warning! Flash has more than one sector size. Remember to create one flash memory region for each sector size range.{colorama.Style.RESET_ALL}")

        data_dict = {
            'name': os.path.splitext(os.path.split(args.elf_path)[-1])[0],
            'prog_header': BLOB_HEADER,
            'header_size': HEADER_SIZE,
            'entry': args.blob_start,
            'stack_base': stack_base,
            'stack_pointer': sp,
            'page_buffers': page_buffers,
            'year': datetime.now().year,
            'copyright_owner': args.copyright or "PyOCD Authors",
        }

        text = algo.process_template(tmpl, data_dict)

        with open(args.output, "w") as file_handle:
            file_handle.write(text)

        print(f"Wrote flash algo dict to {args.output}")

if __name__ == '__main__':
    main()



# pyOCD debugger
# Copyright (c) 2015-2020 Arm Limited
# Copyright (c) 2021-2022 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import print_function

import logging
import os
import sys
import traceback
from xml.etree import ElementTree
import six
import subprocess
import tempfile
import threading
import codecs
from pathlib import Path
from pyocd.utility.compatibility import to_str_safe

OBJCOPY = "arm-none-eabi-objcopy"

TEST_DIR = os.path.dirname(os.path.abspath(__file__))
PYOCD_DIR = os.path.dirname(TEST_DIR)
TEST_DATA_DIR = os.path.join(TEST_DIR, "data")
TEST_OUTPUT_DIR = os.path.join(TEST_DIR, "output")

def get_test_binary_path(binary_name):
    if binary_name is None:
        binary_name = os.environ.get('PYOCD_TEST_BINARY')
        if binary_name is None:
            raise RuntimeError("no test binary available")
    if Path(binary_name).is_absolute():
        return binary_name
    else:
        return os.path.join(TEST_DATA_DIR, "binaries", binary_name)

def get_env_name():
    return os.environ.get('TOX_ENV_NAME', '')

def get_env_file_name():
    env_name = get_env_name()
    return ("_" + env_name) if env_name else ''

def ensure_output_dir():
    if not os.path.isdir(TEST_OUTPUT_DIR):
        if os.path.exists(TEST_OUTPUT_DIR):
            raise RuntimeError("path '%s' already exists but is not a directory" % TEST_OUTPUT_DIR)
        os.mkdir(TEST_OUTPUT_DIR)

# Returns common option values passed in when creating test sessions.
def get_session_options():
    return {
        # These options can be overridden by probe config in pyocd.yaml.
        'option_defaults': {
            'frequency': 1000000, # 1 MHz
            'skip_test': False,
            },
        }

# Returns a dict containing some test parameters for the target in the passed-in session.
#
# 'test_clock' : the max supported SWD frequency for the target
# 'error_on_invalid_access' : whether invalid accesses cause a fault
#
def get_target_test_params(session):
    target_type = session.board.target_type
    error_on_invalid_access = True
    if target_type.startswith("nrf"):
        # Override clock since 10MHz is too fast
        test_clock = 1000000
        error_on_invalid_access = False
    elif target_type == "ncs36510":
        # Override clock since 10MHz is too fast
        test_clock = 1000000
    else:
        # Default of 4 MHz.
        test_clock = 4000000
    return {
            'test_clock': test_clock,
            'error_on_invalid_access': error_on_invalid_access,
            }

# Generate an Intel hex file from the binary test file.
def binary_to_hex_file(binary_file, base_address):
    temp_test_hex_name = tempfile.mktemp('.hex')
    objcopyOutput = subprocess.check_output([OBJCOPY,
        "-v", "-I", "binary", "-O", "ihex", "-B", "arm", "-S",
        "--set-start", "0x%x" % base_address,
        "--change-addresses", "0x%x" % base_address,
        binary_file, temp_test_hex_name], stderr=subprocess.STDOUT)
    print(to_str_safe(objcopyOutput))
    # Need to escape backslashes on Windows.
    if sys.platform.startswith('win'):
        temp_test_hex_name = temp_test_hex_name.replace('\\', '\\\\')
    return temp_test_hex_name

# Generate an elf from the binary test file.
def binary_to_elf_file(binary_file, base_address):
    temp_test_elf_name = tempfile.mktemp('.elf')
    objcopyOutput = subprocess.check_output([OBJCOPY,
        "-v", "-I", "binary", "-O", "elf32-littlearm", "-B", "arm", "-S",
        "--set-start", "0x%x" % base_address,
        "--change-addresses", "0x%x" % base_address,
        binary_file, temp_test_elf_name], stderr=subprocess.STDOUT)
    print(to_str_safe(objcopyOutput))
    # Need to escape backslashes on Windows.
    if sys.platform.startswith('win'):
        temp_test_elf_name = temp_test_elf_name.replace('\\', '\\\\')
    return temp_test_elf_name

def run_in_parallel(function, args_list):
    """Create and run a thread in parallel for each element in args_list

    Wait until all threads finish executing. Throw an exception if an exception
    occurred on any of the threads.
    """
    def _thread_helper(idx, func, args):
        """Run the function and set result to True if there was not error"""
        func(*args)
        result_list[idx] = True

    result_list = [False] * len(args_list)
    thread_list = []
    for idx, args in enumerate(args_list):
        thread = threading.Thread(target=_thread_helper,
                                  args=(idx, function, args))
        thread.start()
        thread_list.append(thread)

    for thread in thread_list:
        thread.join()
    for result in result_list:
        if result is not True:
            raise Exception("Running in thread failed")

def wait_with_deadline(process, timeout):
    try:
        from subprocess import TimeoutExpired
        try:
            process.wait(timeout=timeout)
        except TimeoutExpired as e:
            print('Timeout while waiting for process %s to exit: %s' % (process, e))
            process.kill()
            return False
    except ImportError:
        # Python 2.7 doesn't support deadline for wait.
        # Let's wait without deadline, as Python 2.7 support is close to end anyway.
        process.wait()
    return True

class IOTee(object):
    def __init__(self, *args):
        self.outputs = list(args)

    def add(self, output):
        self.outputs.append(output)

    def write(self, message):
        for out in self.outputs:
            encoding = out.encoding if getattr(out, 'encoding', None) else 'latin-1'

            try:
                # Pre-encode the output with error replacement, then convert back to a string
                # to write to the file. Very inefficient, but prevents possible errors.
                b = codecs.encode(message, encoding=encoding, errors='backslashreplace')
                u = codecs.decode(b, encoding=encoding, errors='backslashreplace')
                out.write(u)
            except UnicodeEncodeError as err:
                out.write(f"<encode error: {err}>")

    def flush(self):
        for out in self.outputs:
            out.flush()

    def isatty(self):
        return False

class RecordingLogHandler(logging.Handler):
    def __init__(self, iostream, level=logging.NOTSET):
        super(RecordingLogHandler, self).__init__(level)
        self.stream = iostream

    def emit(self, record):
        try:
            message = self.format(record)
            self.stream.write(six.u(message + "\n"))
        except Exception:
            self.handleError(record)

class TestResult(object):

    def __init__(self, test_board, test, result):
        self.passed = result
        self._board = test_board.target_type if test_board else 'unknown'
        self.board_name = test_board.name if test_board else ""
        self.test = test
        self.name = "test"
        self.time = 0
        self.output = ""

    @property
    def board(self):
        return self._board

    @board.setter
    def board(self, newBoard):
        self._board = newBoard.target_type if newBoard else 'unknown'
        self.board_name = newBoard.name

    def get_test_case(self):
        if 'TOX_ENV_NAME' in os.environ:
            classname = "{}.{}.{}.{}".format(os.environ['TOX_ENV_NAME'], self.board_name, self.board, self.name)
        else:
            classname = "{}.{}.{}".format(self.board_name, self.board, self.name)
        case = ElementTree.Element('testcase',
                    name=classname,
                    classname=classname,
                    status=("passed" if self.passed else "failed"),
                    time="%.3f" % self.time
                    )
        case.text = "\n"
        case.tail = "\n"
        if not self.passed:
            failed = ElementTree.SubElement(case, 'failure',
                        message="failure",
                        type="failure"
                        )
            failed.text = self.filter_output(self.output)
        system_out = ElementTree.SubElement(case, 'system-out')
        system_out.text = self.filter_output(self.output)
        return case

    def filter_output(self, output):
        """@brief Hex-encode null byte and control characters."""
        result = six.text_type()
        for c in output:
            if (c not in ('\n', '\r', '\t')) and (0 <= ord(c) <= 31):
                result += u"\\x{:02x}".format(ord(c))
            else:
                result += c
        return result

class Test(object):

    def __init__(self, name, function):
        self.name = name
        self.test_function = function
        self.n = 0

    def run(self, board):
        """
        Run test and return the result

        Override this function to return a custom result
        """
        passed = False
        try:
            self.test_function(board.unique_id)
            passed = True
        except Exception as e:
            print("Exception %s when testing board %s" % (e, board.unique_id))
            traceback.print_exc(file=sys.stdout)
        result = TestResult(board, self, passed)
        result.name = self.name
        return result

    def print_perf_info(self, result_list, output_file=None):
        """
        Print performance info if any
        """
        pass

    @staticmethod
    def print_results(result_list, output_file=None, ignored=[]):
        msg_format_str = "{:<15}{:<21}{:<15}{:<15}"
        print("\n\n------ TEST RESULTS ------")
        print(msg_format_str .format("Target", "Test", "Result", "Time"),
              file=output_file)
        print("", file=output_file)
        for result in result_list:
            status_str = "Pass" if result.passed else "Fail"
            if not result.passed and result.test.name in ignored:
                status_str += " [ignored]"
            print(msg_format_str.format(result.board,
                                        result.test.name,
                                        status_str, "%.3f" % result.time),
                  file=output_file)

    @staticmethod
    def all_tests_pass(result_list, ignored=[]):
        passed = True
        for result in result_list:
            if not result.passed and result.test.name not in ignored:
                passed = False
                break
        if len(result_list) <= 0:
            passed = False
        return passed



# pyOCD debugger
# Copyright (c) 2025 Arm Limited
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sysconfig

def get_site_packages() -> str:
    """Get the path to the site-packages directory."""
    path = sysconfig.get_path('purelib')
    if not path:
        raise RuntimeError("Could not determine site-packages directory")
    return path

def main() -> None:
    site_packages = get_site_packages()
    # Check if running in GitHub Actions
    github_env = os.getenv('GITHUB_ENV')

    if github_env:
        # If running in GitHub Actions, write to GITHUB_ENV
        with open(github_env, "a") as env_file:
            env_file.write(f"SITE_PACKAGES={site_packages}\n")
    else:
        # If not running in GitHub Actions, print the path
        print(f"SITE_PACKAGES={site_packages}")

if __name__ == "__main__":
    main()



#!/usr/bin/env python3

# Copyright (c) 2021 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pyocd.core.soc_target import SoCTarget
from pyocd.target.builtin import BUILTIN_TARGETS


def gen_one_target(name: str, target: type[SoCTarget]) -> None:
    print(f"""    <tr><td><code>{name.lower()}</code></td>
    <td>{target.VENDOR}</td>
    <td>{target.__name__}</td>
    </tr>
""")

def gen_targets() -> None:
    for target_name in sorted(BUILTIN_TARGETS.keys()):
        target = BUILTIN_TARGETS[target_name]
        gen_one_target(target_name, target)
        

def main() -> None:
    print("""---
title: Built-in targets
---

<table>

<tr><th>Target Type Name</th><th>Vendor</th><th>Name</th></tr>
""")
    gen_targets()
    print("""
</table>
""")


if __name__ == '__main__':
    main()





# Copyright (c) 2010 - 2023, Nordic Semiconductor ASA All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ...core.memory_map import FlashRegion, RamRegion, MemoryMap
from ...debug.svd.loader import SVDFile
from ..family.target_nRF54L import NRF54L

FLASH_ALGO = {
    'load_address' : 0x20000000,

    # Flash algorithm as a hex string
    'instructions': [
    0xE00ABE00,
    0xf8d24a02, 0x2b013400, 0x4770d1fb, 0x5004b000, 0x47702000, 0x47702000, 0x49072001, 0xf8c1b508,
    0xf7ff0500, 0xf8c1ffed, 0x20000540, 0xffe8f7ff, 0x0500f8c1, 0xbf00bd08, 0x5004b000, 0x2301b508,
    0xf8c14906, 0xf7ff3500, 0xf04fffdb, 0x600333ff, 0xf7ff2000, 0xf8c1ffd5, 0xbd080500, 0x5004b000,
    0x2301b538, 0x4d0c4614, 0x0103f021, 0x3500f8c5, 0xffc6f7ff, 0x44214622, 0x42911b00, 0x2000d105,
    0xffbef7ff, 0x0500f8c5, 0x4613bd38, 0x4b04f853, 0x461a5014, 0xbf00e7f1, 0x5004b000, 0x00000000
    ],

    # Relative function addresses
    'pc_init': 0x20000015,
    'pc_unInit': 0x20000019,
    'pc_program_page': 0x20000065,
    'pc_erase_sector': 0x20000041,
    'pc_eraseAll': 0x2000001d,

    'static_base' : 0x20000000 + 0x00000004 + 0x000000a0,
    'begin_stack' : 0x20000300,
    'begin_data' : 0x20000000 + 0x1000,
    'page_size' : 0x4,
    'analyzer_supported' : False,
    'analyzer_address' : 0x00000000,
    'page_buffers' : [0x20001000, 0x20001004],   # Enable double buffering
    'min_program_length' : 0x4,

    # Flash information
    'flash_start': 0x0,
    'flash_size': 0x17d000,
    'sector_sizes': (
        (0x0, 0x17d000),
        (0xffd000, 0x1000),
    )
}

class NRF54L15(NRF54L):
    MEMORY_MAP = MemoryMap(
        FlashRegion(
            start=0x0,
            length=0x0017D000,
            blocksize=0x1000,
            is_boot_memory=True,
            algo=FLASH_ALGO,
        ),
        # User Information Configation Registers (UICR) as a flash region
        FlashRegion(
            start=0x00FFD000,
            length=0x1000,
            blocksize=0x4,
            is_testable=False,
            is_erasable=False,
            algo=FLASH_ALGO,
        ),
        RamRegion(start=0x20000000, length=0x40000),
    )

    def __init__(self, session):
        super(NRF54L15, self).__init__(session, self.MEMORY_MAP)
        self._svd_location = SVDFile.from_builtin("nrf54l15.svd")



# pyOCD debugger
# Copyright (c) 2021 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from ...core import exceptions
from ...core.target import Target
from ...coresight.coresight_target import CoreSightTarget
from ...core.memory_map import (RomRegion, FlashRegion, RamRegion, MemoryMap)
from ...probe.swj import SWJSequenceSender
from ...probe.debug_probe import DebugProbe
from ...utility import mask

LOG = logging.getLogger(__name__)

FLASH_ALGO = {
    'load_address' : 0x20000000,

    # Flash algorithm as a hex string
    'instructions': [
    0xe7fdbe00,
    0xaf03b5f0, 0x4614b087, 0x447d4d5a, 0x28007828, 0x485ed007, 0x68004478, 0x485d4780, 0x68004478,
    0x26014780, 0x1e60702e, 0xd3002803, 0x4848e08d, 0xf90ef000, 0x2800460c, 0x4846d15f, 0xf0001c80,
    0x2800f907, 0x460cd001, 0x9406e057, 0x78102210, 0x284d4c41, 0x2311d151, 0x28757818, 0x2012d14d,
    0x78009002, 0xd00a2802, 0xd1462801, 0x92049303, 0x20149105, 0x21188800, 0x4938880a, 0x9303e008,
    0x91059204, 0xf8d4f000, 0x88022016, 0x21044833, 0x4c334790, 0x46222800, 0x4602d000, 0x99052800,
    0x9b039804, 0x7800d029, 0x284d4c2e, 0x7818d125, 0xd1222875, 0x78009802, 0x91052802, 0xd0079201,
    0xd11a2801, 0x88002014, 0x880a2118, 0xe0054926, 0xf8aef000, 0x88022016, 0x21044823, 0x4c234790,
    0x46222800, 0x4602d000, 0xd0062800, 0x48209204, 0xf8aef000, 0x2800460c, 0x4620d002, 0xbdf0b007,
    0xf0004814, 0x2800f8a5, 0x9103d19d, 0x28009806, 0x9806d019, 0x98054780, 0x48174780, 0x99014478,
    0x48166001, 0x99064478, 0x48156001, 0x99044478, 0x48146001, 0x60044478, 0x44784813, 0x60019903,
    0x2400702e, 0x9c05e7d9, 0xf000e7d7, 0x46c0f8ab, 0x00004649, 0x00005843, 0x10004552, 0x00004552,
    0x20004552, 0x10005052, 0x00005052, 0x20005052, 0x00004346, 0x0000029e, 0x00000194, 0x00000188,
    0x00000188, 0x00000184, 0x00000182, 0x000002a4, 0x000002a0, 0xaf02b5d0, 0x447c4c08, 0x28017820,
    0x4807d10a, 0x68004478, 0x48064780, 0x68004478, 0x20004780, 0xbdd07020, 0xbdd02001, 0x0000010e,
    0x00000114, 0x00000110, 0xaf02b5d0, 0x44794909, 0x29017809, 0x210fd10c, 0x18400709, 0x44794906,
    0x2201680c, 0x04120311, 0x47a023d8, 0xbdd02000, 0xbdd02001, 0x000000da, 0x000000d2, 0xaf02b5d0,
    0x4909460b, 0x78094479, 0xd10a2901, 0x0709210f, 0x49061840, 0x680c4479, 0x461a4611, 0x200047a0,
    0x2001bdd0, 0x46c0bdd0, 0x000000a4, 0x000000a0, 0xf45f4806, 0x60014140, 0xf710ee30, 0xec40d404,
    0xec400780, 0xbf400781, 0x00004770, 0xe000ed88, 0xaf02b5d0, 0x2010b284, 0x284d7800, 0x2011d10f,
    0x28757800, 0x2012d10b, 0x28027800, 0x2801d00b, 0x2014d105, 0x21188800, 0x4621880a, 0x2001e009,
    0x18610701, 0xf7ffbdd0, 0x2016ffd3, 0x21048802, 0x47904620, 0x42404601, 0x29004148, 0x2101d1f2,
    0xe7ee0749, 0xaf00b580, 0xdefede00, 0xd4d4d400, 0x00000000, 0x00000000, 0x00000000, 0x00000000,
    0x00000000
    ],

    # Relative function addresses
    'pc_init': 0x20000005,
    'pc_unInit': 0x20000199,
    'pc_program_page': 0x20000201,
    'pc_erase_sector': 0x200001cd,
    'pc_eraseAll': 0x120000003,

    'static_base' : 0x20000000 + 0x00000004 + 0x000002c4,
    'begin_stack' : 0x200012d0,
    'end_stack' : 0x200002d0,
    'begin_data' : 0x20000000 + 0x1000,
    'page_size' : 0x0,
    'analyzer_supported' : False,
    'page_buffers' : [0x20001400, 0x20001500],   # Enable double buffering
    'min_program_length' : 256,

    # Relative region addresses and sizes
    'ro_start': 0x4,
    'ro_size': 0x2c4,
    'rw_start': 0x2c8,
    'rw_size': 0x0,
    'zi_start': 0x2c8,
    'zi_size': 0x0,
}

def _parity32(value):
    parity = sum((value >> i) for i in range(32))
    return parity & 1

class RP2Base(CoreSightTarget):
    VENDOR = "Raspberry Pi"

    MEMORY_MAP = MemoryMap(
        RomRegion(  start=0,            length=0x4000,      name="bootrom",             ),
        FlashRegion(start=0x10000000,   length=0x1000000,   name="xip",
            sector_size=4096,
            page_size=256,
            algo=FLASH_ALGO,
            is_boot_memory=True),
        RomRegion(  start=0x11000000,   length=0x1000000,   name="xip_noalloc",           alias="xip"),
        RomRegion(  start=0x12000000,   length=0x1000000,   name="xip_nocache",           alias="xip"),
        RomRegion(  start=0x13000000,   length=0x1000000,   name="xip_noalloc_nocache",   alias="xip"),
        RamRegion(  start=0x20000000,   length=0x40000,     name="sram0_3",               ),
        RamRegion(  start=0x20040000,   length=0x2000,      name="sram4_5",               ),
        RamRegion(  start=0x21000000,   length=0x40000,     name="sram0_3_alias",         alias="sram0_3"),
        RamRegion(  start=0x51000000,   length=0x1000,      name="usbram",                ),
        )

    def __init__(self, session):
        super().__init__(session, self.MEMORY_MAP)


class RP2350(RP2Base):
    """@brief Raspberry Pi RP2350
    """

    def __init__(self, session):
        super().__init__(session)

    def create_init_sequence(self):
        seq = super().create_init_sequence()

        # Secure mode is only needed for the flash algo.
        seq.insert_before('post_connect_hook', ('set_set_secure_mode', self._set_secure_mode))

        return seq

    def _set_secure_mode(self):
        # The RP2350 flash functions in ROM require the core to be in secure mode
        state = self.session.board.target.get_security_state()
        if state == Target.SecurityState.SECURE:
            LOG.debug("target in secure mode")
            return

        LOG.debug("target not in secure mode, attempting to switch to secure mode")
        target = self.session.board.target

        DCB_DSCSR = 0xE000EE08
        DSCSR_CDSKEY = 1 << 17
        DSCSR_CDS = 1 << 16
        dscsr = target.read32(DCB_DSCSR)
        target.write32(DCB_DSCSR, (dscsr & ~DSCSR_CDSKEY) | DSCSR_CDS)

        state = self.session.board.target.get_security_state()
        if state != Target.SecurityState.SECURE:
            LOG.debug("target failed to enter secure mode")
            raise exceptions.TargetError("Unable to set target to secure mode")

        # Attempt to enable secure access to SRAM
        ACCESSCTRL_LOCK =  0x40060000
        ACCESSCTRL_LOCK_DEBUG_BITS = 0x00000008

        ACCESSCTRL_CFGRESET = 0x40060008
        ACCESSCTRL_WRITE_PASSWORD = 0xacce0000
        lock = target.read32(ACCESSCTRL_LOCK)
        if lock & ACCESSCTRL_LOCK_DEBUG_BITS:
            # Warn instead of rasing an error in case the permissions are setup
            # correctly.
            LOG.warn("ACCESSCTRL is locked.  Unable to reset.")
        else:
            target.write32(ACCESSCTRL_CFGRESET, ACCESSCTRL_WRITE_PASSWORD | 1)


class RP2040Base(RP2Base):
    """@brief Raspberry Pi RP2040.

    This device is very strange in that it as three DPs. The first two DPs each have a single AHB-AP
    for the two Cortex-M0+ cores. The third DP is a "Rescue DP" that has no APs, but the CDBGPWRUPREQ
    signal is repurposed as a rescue signal.
    """

    class Targetsel:
        """@brief DP TARGETEL values for each DP."""
        CORE_0 = 0x01002927
        CORE_1 = 0x11002927
        RESCUE_DP = 0xf1002927

    def __init__(self, session):
        super().__init__(session)

        ## The TARGETSEL value to be used.
        self._core_targetsel = None

        # Disable the SWJ sequence. This performs a line reset, which causes all DPs to be selected. After
        # any line reset on a multi-drop SWD setup there must be a target selection sequence.
        session.options['dap_swj_enable'] = False

    def create_init_sequence(self):
        seq = super().create_init_sequence()

        seq.insert_before('load_svd', ('check_probe', self._check_probe)) \
            .insert_before('dp_init', ('select_core0', self._select_core))

        return seq

    def _check_probe(self):
        # Have to import here to avoid a circular import
        from ...probe.debug_probe import DebugProbe
        if DebugProbe.Capability.SWD_SEQUENCE not in self.session.probe.capabilities:
            raise exceptions.TargetSupportError("RP2040 requires a debug probe with SWD sequence capability")

    def _select_core(self):
        self.select_dp(self._core_targetsel)

    def select_dp(self, targetsel):
        """@brief Select the DP with the matching TARGETSEL."""
        probe = self.session.probe

        # Have to connect the probe first, or SWCLK will not be enabled.
        probe.connect(DebugProbe.Protocol.SWD)

        # First perform the dormant to SWD sequence. An SW-DPv2 implementing multi-drop SWD will cold
        # reset into dormant mode.
        swj = SWJSequenceSender(self.session.probe, True)
        swj.dormant_to_swd()

        # SWD line reset to activate all DPs.
        swj.line_reset()
        swj.idle_cycles(2)

        # Send multi-drop SWD target selection sequence to select the requested DP.
        probe.swd_sequence([
            # DP TARGETSEL write
            # output 8 cycles:
            #   - Start = 1
            #   - APnDP = 0
            #   - RnW = 0
            #   - A[2:3] = 2'b11
            #   - Parity = 0
            #   - Stop = 0
            #   - Park = 1
            # -> LSB first, that's 0b10011001 or 0x99
            (8, 0x99),

            # 5 cycles with SWDIO as input
            (5,),

            # DP TARGETSEL value
            # output 32 + 1 cycles
            (33, targetsel | mask.parity32_high(targetsel)),

            # 2 idle cycles
            (2, 0x00),
            ])

        DP_IDR = 0x00
        dpidr = probe.read_dp(DP_IDR)
        LOG.debug("DP IDR after writing TARGETSEL: 0x%08x", dpidr)

        probe.write_dp(0x8, 0x2) # DPBANKSEL=2 to select TARGETID
        targetid = probe.read_dp(0x4)
        LOG.debug("DP TARGETID: 0x%08x", targetid)

        probe.write_dp(0x8, 0x3) # DPBANKSEL=3 to select DLPIDR
        dlpidr = probe.read_dp(0x4)
        LOG.debug("DP DLPIDR: 0x%08x", dlpidr)

        probe.write_dp(0x8, 0x0) # restore DPBANKSEL=0

class RP2040Core0(RP2040Base):
    """@brief RP2040 target for core 0."""

    def __init__(self, session):
        super().__init__(session)
        self._core_targetsel = self.Targetsel.CORE_0

class RP2040Core1(RP2040Base):
    """@brief RP2040 target for core 1."""

    def __init__(self, session):
        super().__init__(session)
        self._core_targetsel = self.Targetsel.CORE_1
